from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestBallDevice(MpfTestCase):
    def __init__(self, test_map):
        super().__init__(test_map)
        self._captured = 0
        self._enter = -1
        self._missing = 0
        self._requesting = 0
        self._queue = False

    def get_config_file(self):
        return 'test_ball_device.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/ball_device/'

    def _missing_ball(self, **kwargs):
        del kwargs
        self._missing += 1

    def test_placeholder(self):
        template = self.machine.placeholder_manager.build_int_template(
            "device.ball_devices.test_launcher.balls", None)
        value, future = template.evaluate_and_subscribe([])
        self.assertEqual(0, value)
        self.assertFalse(future.done())
        self.hit_switch_and_run("s_ball_switch_launcher", 1)
        self.assertTrue(future.done())

    def test_ball_count_during_eject(self):
        coil2 = self.machine.coils['eject_coil2']
        device2 = self.machine.ball_devices['test_launcher']
        coil2.pulse = MagicMock()

        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)

        self._missing = 0
        self.assertEqual(0, self.machine.playfield.balls)
        self.assertEqual(0, self.machine.playfield.available_balls)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)
        self.assertEqual(1, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, self.machine.playfield.balls)
        self.assertEqual(1, self.machine.playfield.available_balls)

        self.assertTrue(coil2.pulse.called)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)

        self.assertEqual(0, self._missing)
        self.advance_time_and_run(300)
        self.assertEqual(1, self._missing)
        self.assertEqual(1, self.machine.playfield.balls)
        self.assertEqual(1, self.machine.playfield.available_balls)

    def _requesting_ball(self, balls, **kwargs):
        del kwargs
        self._requesting += balls

    def test_ball_eject_failed(self):
        self._requesting = 0
        coil2 = self.machine.coils['eject_coil2']
        coil2.pulse = MagicMock()

        self.machine.events.add_handler(
            'balldevice_test_launcher_ball_request', self._requesting_ball)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        # launcher should eject
        self.advance_time_and_run(1)
        self.assertTrue(coil2.pulse.called)

        self._captured = 0

        # launcher should retry eject
        self.advance_time_and_run(10)
        self.assertEqual(2, coil2.pulse.call_count)

        self.assertEqual(0, self._requesting)

        # it should not claim a ball which enters the target
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)

        self.advance_time_and_run(300)

    def test_ball_eject_timeout_and_late_confirm(self):
        self._requesting = 0
        coil2 = self.machine.coils['eject_coil2']
        coil2.pulse = MagicMock()
        self._missing = 0

        self.machine.events.add_handler(
            'balldevice_test_launcher_ball_request', self._requesting_ball)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        # launcher should eject
        self.advance_time_and_run(1)
        self.assertTrue(coil2.pulse.called)
        self._captured = 0

        # it leaves the switch
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        # but not confirm. eject timeout = 6s
        self.advance_time_and_run(15)
        self.assertTrue(coil2.pulse.called)

        # late confirm
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)

        self.assertEqual(0, self._requesting)
        self.assertEqual(0, self._missing)
        self.assertEqual(0, self._captured)

    def test_ball_left_and_return_failure(self):
        self._requesting = 0
        coil2 = self.machine.coils['eject_coil2']
        coil2.pulse = MagicMock()
        self._missing = 0

        self.machine.events.add_handler(
            'balldevice_test_launcher_ball_request', self._requesting_ball)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        # launcher should eject
        self.advance_time_and_run(1)
        self.assertTrue(coil2.pulse.called)
        self._captured = 0

        # it leaves the switch
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertTrue(coil2.pulse.called)

        # it comes back (before timeout)
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertTrue(coil2.pulse.called)

        # retry after timeout
        self.advance_time_and_run(5)
        self.assertEqual(2, coil2.pulse.call_count)

        self.assertEqual(0, self._captured)

        # ball did not leave the launcher (it returned). target should capture
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)

        self.assertEqual(0, self._requesting)
        self.assertEqual(0, self._missing)
        self.assertEqual(1, self._captured)

    def test_eject_retry(self):
        self.hit_switch_and_run("s_ball_switch1", 1)
        self.hit_switch_and_run("s_ball_switch2", 1)
        self.assertAvailableBallsOnPlayfield(0)

        coil2 = self.machine.coils['eject_coil2'].hw_driver

        self.mock_event('balldevice_test_launcher_ball_request')
        self.mock_event('balldevice_ball_missing')
        self.mock_event('balldevice_captured_from_playfield')

        # launcher should eject
        self.hit_switch_and_run("s_ball_switch_launcher", 1)
        self.assertEqual("pulsed_10", coil2.state)
        coil2.state = None
        self.assertEventCalled('balldevice_captured_from_playfield')
        self.mock_event('balldevice_captured_from_playfield')

        # it leaves the switch
        self.release_switch_and_run("s_ball_switch_launcher", 1)
        self.assertEqual(None, coil2.state)

        # it comes back (before timeout)
        self.hit_switch_and_run("s_ball_switch_launcher", 1)
        self.assertEqual(None, coil2.state)

        # retry after timeout
        self.advance_time_and_run(5)
        self.assertEqual("pulsed_10", coil2.state)
        coil2.state = None

        # ball leaves launcher
        self.release_switch_and_run("s_ball_switch_launcher", 1)
        self.advance_time_and_run(1)

        # arrives at target
        self.hit_switch_and_run("s_ball_switch_target1", 1)

        self.assertEqual(None, coil2.state)

        self.assertEventNotCalled('balldevice_captured_from_playfield')
        self.assertEventNotCalled('balldevice_ball_missing')
        self.assertEventNotCalled('balldevice_test_launcher_ball_request')
        self.assertAvailableBallsOnPlayfield(1)
        self.assertBallsOnPlayfield(0)

    def test_ball_eject_timeout_and_ball_missing(self):
        self._requesting = 0
        coil2 = self.machine.coils['eject_coil2']
        playfield = self.machine.ball_devices['playfield']
        coil2.pulse = MagicMock()
        self._missing = 0
        self._captured = 0
        self._enter = 0

        self.mock_event("balldevice_test_launcher_ball_eject_failed")
        self.mock_event("balldevice_test_launcher_ball_eject_success")

        self.machine.events.add_handler(
            'balldevice_test_launcher_ball_request', self._requesting_ball)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_test_target1_ball_enter',
                                        self._ball_enter)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        # launcher should eject
        self.advance_time_and_run(1)
        self.assertTrue(coil2.pulse.called)
        self.assertEqual(1, self._captured)

        # it leaves the switch
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)

        # but not confirm. eject timeout = 6s
        self.advance_time_and_run(15)
        self.assertTrue(coil2.pulse.called)

        self.advance_time_and_run(30)

        self.assertEqual(0, self._requesting)
        self.assertEqual(1, self._missing)
        self.assertEqual(1, playfield.balls)

        self.assertEventNotCalled("balldevice_test_launcher_ball_eject_success")
        self.assertEventCalled("balldevice_test_launcher_ball_eject_failed")

        self._missing = 0
        self._requesting = 0
        self._captured = 0

        # target1 captures a ball since the eject failed
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)

        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual(0, playfield.balls)

        # and ejects it
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, playfield.balls)

        self._captured = 0

        # launcher captures a ball and should retry
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(2, coil2.pulse.call_count)

        self.assertEqual(1, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual(0, playfield.balls)

        self._captured = 0

        # it leaves the switch
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)

        # and reaches target which claims it
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual(0, playfield.balls)

    def test_eject_successful_to_playfield(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device3 = self.machine.ball_devices['test_target1']
        playfield = self.machine.ball_devices['playfield']

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # request an ball
        playfield.add_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # target1 receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device3.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        self.assertTrue(coil3.pulse.called)
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device3.balls)

        # a ball hits a playfield switch
        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(1, playfield.balls)

    def _ball_enter(self, new_balls, unclaimed_balls, **kwargs):
        del kwargs
        del new_balls
        if unclaimed_balls < 0:
            raise Exception("Balls went negative")

        self._enter += unclaimed_balls

    def _captured_from_pf(self, balls, **kwargs):
        del kwargs
        self._captured += balls

    def test_eject_successful_to_other_trough(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device4 = self.machine.ball_devices['test_target2']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_test_target2_ball_enter',
                                        self._ball_enter)
        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._enter = -1
        self._captured = 0
        self._missing = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = -1

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(1, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # request an ball
        device4.request_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # target2 receives and keeps ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual(-1, self._enter)
        self.assertEqual(-1, self._captured)

        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, self._missing)

    def test_eject_to_pf_and_other_trough(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device3 = self.machine.ball_devices['test_target1']
        device4 = self.machine.ball_devices['test_target2']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = -1

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # request ball
        device4.request_ball()
        self.advance_time_and_run(1)

        # request an ball
        playfield.add_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # target2 receives and keeps ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        # eject of launcher should be confirmed now and the trough should eject
        self.assertEqual(2, coil1.pulse.call_count)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        self.assertEqual(-1, self._captured)

        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, self._missing)

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertEqual(2, coil1.pulse.call_count)
        self.assertEqual(2, coil2.pulse.call_count)
        assert not coil3.pulse.called
        assert not coil4.pulse.called

        # ball leaves launcher
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)

        # target1 receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device3.balls)

        self.assertEqual(2, coil1.pulse.call_count)
        self.assertEqual(2, coil2.pulse.call_count)
        self.assertTrue(coil3.pulse.called)
        assert not coil4.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device3.balls)

        # a ball hits a playfield switch
        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        self.assertEqual(-1, self._captured)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._missing)

    def test_eject_ok_to_receive(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device3 = self.machine.ball_devices['test_target1']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = -1

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # request an ball to pf
        playfield.add_ball()
        self.advance_time_and_run(1)

        # request a second ball to pf
        playfield.add_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()

        # target1 receives and should eject it right away
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device3.balls)

        # eject of launcher should be confirmed now and trough can eject
        # the next ball
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        self.assertTrue(coil3.pulse.called)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # launcher receives a ball but cannot send it to target1 because its busy
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        self.assertTrue(coil3.pulse.called)

        self.assertEqual(-1, self._captured)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, self._missing)

        # ball left target1
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device3.balls)

        # wait for confirm
        self.advance_time_and_run(10)

        # launcher should now eject the second ball
        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        self.assertTrue(coil3.pulse.called)

        # ball leaves launcher
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)

        # target1 receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device3.balls)

        self.assertEqual(1, coil1.pulse.call_count)
        self.assertEqual(1, coil2.pulse.call_count)
        self.assertEqual(2, coil3.pulse.call_count)

        # ball left target1
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device3.balls)

        # wait for confirm
        self.advance_time_and_run(10)

        self.assertEqual(-1, self._captured)

        self.assertEqual(2, playfield.balls)
        self.assertEqual(0, self._missing)

        # check that timeout behave well
        self.advance_time_and_run(300)

    def test_temporary_missing_ball_idle(self):
        coil1 = self.machine.coils['eject_coil1']
        device1 = self.machine.ball_devices['test_trough']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        self.assertEqual(0, playfield.balls)

        # it should keep the balls
        coil1.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        self.assertEqual(2, device1.available_balls)

        # steal a ball from trough (only for a while)
        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(3)
        assert not coil1.pulse.called
        self.assertEqual(0, self._missing)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)

        # still waiting
        self.assertEqual(2, device1.balls)
        self.assertEqual(2, device1.available_balls)

        # put it back
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        assert not coil1.pulse.called
        self.assertEqual(0, self._missing)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)

        # count should be back
        self.assertEqual(2, device1.balls)
        self.assertEqual(2, device1.available_balls)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # steal a ball from trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(6)
        assert not coil1.pulse.called
        self.assertEqual(1, self._missing)
        self.assertEqual(0, self._captured)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, playfield.available_balls)

        # count should be on less and one ball missing
        self.assertEqual(1, device1.balls)
        self.assertEqual(1, device1.available_balls)

        # put it back
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)
        assert not coil1.pulse.called
        self.assertEqual(1, self._missing)
        self.assertEqual(1, self._captured)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)

        # count should be back
        self.assertEqual(2, device1.balls)
        self.assertEqual(2, device1.available_balls)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

    def test_missing_ball_idle(self):
        coil1 = self.machine.coils['eject_coil1']
        device1 = self.machine.ball_devices['test_trough']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0

        self.assertEqual(0, playfield.balls)

        # it should keep the balls
        coil1.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        self.assertEqual(2, device1.available_balls)

        # steal a ball from trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(6)
        assert not coil1.pulse.called
        self.assertEqual(1, self._missing)
        self.assertEqual(0, self._captured)
        self.assertEqual(1, playfield.balls)
        self.assertEqual(1, playfield.available_balls)

        # count should be on less and one ball missing
        self.assertEqual(1, device1.balls)
        self.assertEqual(1, device1.available_balls)

        # request an ball
        playfield.add_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device1.balls)

        # ball randomly reappears
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, self._missing)
        self.assertEqual(1, self._captured)

    def test_ball_entry_during_eject(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device3 = self.machine.ball_devices['test_target1']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0

        # assume there are more balls in the machine
        self.machine.ball_controller.num_balls_known = 4

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # assume there are already two balls on the playfield
        playfield.balls = 2
        playfield.available_balls = 2

        # request an ball to pf
        playfield.add_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # important: ball does not leave launcher here

        # target1 receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device3.balls)

        self.assertEqual(1, coil1.pulse.call_count)
        self.assertEqual(1, coil2.pulse.call_count)
        self.assertEqual(1, coil3.pulse.call_count)

        # ball left target1
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device3.balls)

        # wait for confirm via timeout
        self.advance_time_and_run(10)

        # target captured one ball because it did not leave the launcher
        self.assertEqual(1, self._captured)

        # there is no new ball on the playfield because the ball is still in the launcher
        self.assertEqual(2, playfield.balls)
        self.assertEqual(0, self._missing)

        # ball disappears from launcher
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # eject times out
        self.advance_time_and_run(15)
        # ball goes missing and magically the playfield count is right again
        self.advance_time_and_run(40)
        self.assertEqual(1, self._missing)
        self.assertEqual(3, playfield.balls)

        # check that timeout behave well
        self.advance_time_and_run(300)

    def _block_eject(self, queue, **kwargs):
        del kwargs
        self._queue = queue
        queue.wait()

    def test_ball_entry_during_ball_requested(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        coil5 = self.machine.coils['eject_coil5']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device4 = self.machine.ball_devices['test_target2']
        target3 = self.machine.ball_devices['test_target3']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        coil5.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called
        assert not coil5.pulse.called

        # request ball
        target3.request_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called
        assert not coil5.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # in the meantime device4 receives a (drained) ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, device4.balls)
        self.assertEqual(2, self._captured)
        self._captured = 0

        # launcher receives but cannot ejects ball yet
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        # target 2 ejects to target 3
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        assert not coil5.pulse.called

        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        # still no eject of launcher
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        assert not coil5.pulse.called

        # target 3 receives
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        # launcher should eject
        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        assert not coil5.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # target2 receives and keeps ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

        self.assertEqual(1, target3.available_balls)

        # request another ball to target3 which already has a ball
        target3.request_ball()
        self.advance_time_and_run(1)
        self.assertEqual(2, target3.available_balls)

    def test_eject_attempt_blocking(self):
        # this test is a bit plastic. we hack get_additional_ball_capacity
        # the launcher device will try to do an eject while device4 is busy.
        # at the moment we cannot trigger this case but it may happen with
        # devices which wait before they eject
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        coil5 = self.machine.coils['eject_coil5']
        device1 = self.machine.ball_devices['test_trough']
        device2 = self.machine.ball_devices['test_launcher']
        device4 = self.machine.ball_devices['test_target2']
        target3 = self.machine.ball_devices['test_target3']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0

        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        coil5.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called
        assert not coil5.pulse.called

        # request ball
        target3.request_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called
        assert not coil5.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # in the meantime target2 receives two (drained) balls
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, device4.balls)
        self.assertEqual(2, self._captured)
        self._captured = 0

        # launcher receives but cannot ejects ball yet
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device2.balls)

        # however launcher will try to eject because we hacked
        # get_additional_ball_capacity. device4 should block the eject

        # target 2 ejects to target 3
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        assert not coil5.pulse.called

        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        # still no eject of launcher
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        assert not coil5.pulse.called

        # target 3 receives
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)

        # launcher should eject
        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        assert not coil5.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, device2.balls)

        # target2 receives and keeps ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, device4.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

    def test_two_concurrent_eject_to_pf_with_no_balls(self):
        coil3 = self.machine.coils['eject_coil3']
        coil5 = self.machine.coils['eject_coil5']
        target1 = self.machine.ball_devices['test_target1']
        target3 = self.machine.ball_devices['test_target3']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)

        coil3.pulse = MagicMock()
        coil5.pulse = MagicMock()

        # add a ball to target1
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        # add a ball to target3
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      1)
        self.advance_time_and_run(1)

        self._captured = 0
        self._missing = 0

        # both should eject to pf
        self.assertEqual(0, playfield.balls)
        self.assertEqual(1, target1.balls)
        self.assertEqual(1, target3.balls)

        self.assertTrue(coil3.pulse.called)
        self.assertTrue(coil5.pulse.called)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        # add a ball to target3
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      0)
        self.advance_time_and_run(1)

        # a ball hits a playfield switch
        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)

        # it should confirm only one device (no matter which one)
        self.assertEqual(1, playfield.balls)

        # second ball should be confirmed via timeout
        self.advance_time_and_run(10)
        self.assertEqual(2, playfield.balls)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)

    def test_unstable_switches(self):
        device1 = self.machine.ball_devices['test_trough']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(0.4)
        # however, second switch is unstable
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(0.4)
        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(0.4)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(0.4)
        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(0.4)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(0.4)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, device1.balls)
        self.assertEqual(0, playfield.balls)
        self.assertRaises(ValueError, device1.ball_count_handler.counter.count_balls_sync)

        # the the other one
        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(0.4)
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(0.4)
        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(0.4)
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(0.4)

        self.assertEqual(0, self._captured)
        self.assertEqual(0, device1.balls)
        self.assertEqual(0, playfield.balls)
        self.assertRaises(ValueError, device1.ball_count_handler.counter.count_balls_sync)

        self.advance_time_and_run(1)
        # but finally both are stable

        self.assertEqual(2, self._captured)
        self.assertEqual(2, device1.balls)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, self._missing)

        self.advance_time_and_run(100)
        self.assertEqual(2, self._captured)
        self.assertEqual(2, device1.balls)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, self._missing)
        #self.assertEqual("idle", device1._state)

    def test_permanent_eject_failure(self):
        coil1 = self.machine.coils['eject_coil1']
        device1 = self.machine.ball_devices['test_trough']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # add two initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)

        # it should keep the ball
        coil1.pulse = MagicMock()
        self.assertEqual(2, device1.balls)
        assert not coil1.pulse.called

        # request ball
        playfield.add_ball()
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        coil1.pulse = MagicMock()

        # timeout is 10s and max 3 retries
        # ball leaves (1st)
        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # and comes back before timeout
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        # after the timeout it should retry
        self.advance_time_and_run(10)
        self.assertTrue(coil1.pulse.called)
        coil1.pulse = MagicMock()

        # ball leaves (2nd) for more than timeout
        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(11)
        self.assertEqual(1, device1.balls)

        # and comes back
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        # trough should retry nearly instantly
        self.advance_time_and_run(1)
        self.assertTrue(coil1.pulse.called)
        coil1.pulse = MagicMock()

        # ball leaves (3rd)
        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, device1.balls)

        # and comes back before timeout
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        # after the timeout the device marks itself as broken and will give up
        self.advance_time_and_run(10)
        assert not coil1.pulse.called

        self.assertEqual(0, self._captured)
        self.assertEqual(0, self._missing)
        self.assertEqual("eject_broken", device1._state)

    def test_request_loops(self):
        # nobody has a ball and we request one. then we add a ball in the chain
        trough = self.machine.ball_devices['test_trough']
        launcher = self.machine.ball_devices['test_launcher']
        target1 = self.machine.ball_devices['test_target1']
        playfield = self.machine.ball_devices['playfield']

        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']
        coil5 = self.machine.coils['eject_coil5']

        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()
        coil5.pulse = MagicMock()

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # no initial balls
        # request ball
        playfield.add_ball()
        self.advance_time_and_run(1)

        self.assertEqual(1, len(target1._ball_requests))
        # should not crash

        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        assert not coil4.pulse.called
        assert not coil5.pulse.called

        # trough captures2 a ball
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0

        self.assertEqual(0, len(target1._ball_requests))

        # trough2 eject
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        assert not coil5.pulse.called
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 0)
        self.advance_time_and_run(1)

        # ball enters launcher2
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      1)
        self.advance_time_and_run(1)

        # launcher2 eject
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        self.assertTrue(coil5.pulse.called)
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      0)
        self.advance_time_and_run(1)

        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)

        # trough eject
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        self.assertTrue(coil5.pulse.called)

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, trough.balls)

        # launcher receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        self.assertTrue(coil4.pulse.called)
        self.assertTrue(coil5.pulse.called)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)

        # target1 receives and ejects ball
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, target1.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        self.assertTrue(coil3.pulse.called)
        self.assertTrue(coil4.pulse.called)
        self.assertTrue(coil5.pulse.called)

        self.assertEqual(0, playfield.balls)
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.advance_time_and_run(10)

        #self.assertEqual("idle", launcher._state)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)

    def test_unexpected_balls(self):
        launcher = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']

        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self._captured = 0
        self._missing = 0

        # launcher catches a random ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, self._captured)
        self._captured = 0

        # trough2 eject
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)

        # ball enters target1
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)

        # target1 should put it to the playfield
        self.assertTrue(coil2.pulse.called)
        self.assertTrue(coil3.pulse.called)

        self.assertEqual(0, playfield.balls)
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      0)
        self.advance_time_and_run(1)
        self.advance_time_and_run(10)

        #self.assertEqual("idle", launcher._state)

        self.assertEqual(1, playfield.balls)
        self.assertEqual(0, self._captured)

    def test_balls_in_device_on_boot(self):
        # The device without the home tag should eject the ball
        # The device with the home tag should not eject the ball

        target1 = self.machine.ball_devices['test_target1']
        target2 = self.machine.ball_devices['test_target2']

        assert 'home' not in target1.tags
        assert 'home' in target2.tags

        # Put balls in both devices
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target2_1", 1)

        coil3 = self.machine.coils['eject_coil3']
        coil4 = self.machine.coils['eject_coil4']

        coil3.pulse = MagicMock()
        coil4.pulse = MagicMock()

        self.advance_time_and_run(10)

        self.assertTrue(coil3.pulse.called)
        assert not coil4.pulse.called

    def _collecting_balls_complete_handler(self, **kwargs):
        del kwargs
        self._collecting_balls_complete = 1

    def test_ball_missing_to_pf_and_drain_no_switches(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        trough = self.machine.ball_devices['test_trough']
        launcher = self.machine.ball_devices['test_launcher']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('collecting_balls_complete',
                                        self._collecting_balls_complete_handler)

        self.assertEqual(None, self.machine.game)

        self._enter = 0
        self._captured = 0
        self._missing = 0
        self._collecting_balls_complete = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, trough.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # start a game
        self.machine.switch_controller.process_switch("s_start", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)

        # there should be a game
        self.assertNotEqual(None, self.machine.game)

        # playfield should expect to have a ball
        self.assertEqual(1, playfield.available_balls)
        self.assertEqual(0, playfield.balls)

        # trough ejects
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(1, playfield.available_balls)
        self.assertEqual(0, playfield.balls)

        # the ball should go to target1. however, it jumps to the pf and drains
        # without hitting a switch on the pf

        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)

        # game should end
        self.assertEqual(None, self.machine.game)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # playfield has negative balls
        self.assertEqual(-1, playfield.balls)
        # but still expects one ball
        self.assertEqual(0, playfield.available_balls)

        self.advance_time_and_run(30)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(0, playfield.available_balls)

    def test_ball_missing_to_pf_and_drain_with_pf_switch(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        trough = self.machine.ball_devices['test_trough']
        launcher = self.machine.ball_devices['test_launcher']
        target = self.machine.ball_devices['test_target1']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('collecting_balls_complete',
                                        self._collecting_balls_complete_handler)

        self.assertEqual(None, self.machine.game)

        self._enter = 0
        self._captured = 0
        self._missing = 0
        self._collecting_balls_complete = 0

        # add an initial ball to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, trough.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # start a game
        self.machine.switch_controller.process_switch("s_start", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_start", 0)
        self.advance_time_and_run(1)

        # there should be a game
        self.assertNotEqual(None, self.machine.game)

        # trough ejects
        self.assertTrue(coil1.pulse.called)
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertTrue(coil1.pulse.called)
        self.assertTrue(coil2.pulse.called)
        assert not coil3.pulse.called

        # launcher shoots the ball
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)

        # the ball should go to target1. however, it jumps to the pf and hits
        # a switch

        self.machine.switch_controller.process_switch("s_playfield", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_playfield", 0)
        self.advance_time_and_run(1)
        self.advance_time_and_run(10)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # player plays a while
        self.advance_time_and_run(40)
        self.advance_time_and_run(40)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, len(target.incoming_balls_handler._incoming_balls))

        # ball drains
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.advance_time_and_run(1)

        # game should end
        self.assertEqual(None, self.machine.game)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        self.advance_time_and_run(30)

        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self.assertEqual(0, playfield.available_balls)

    def test_concurrent_capture_and_eject_unclaimed_balls(self):
        playfield = self.machine.ball_devices['playfield']
        coil5 = self.machine.coils['eject_coil5']
        coil5.pulse = MagicMock()

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self._captured = 0
        self.machine.ball_controller.num_balls_known = 2
        playfield.available_balls = 2
        playfield.balls = 2

        # device captures two balls
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      1)
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target3_2", 1)
        self.advance_time_and_run(.6)
        self.assertEqual(2, self._captured)
        self.assertEqual(0, playfield.balls)
        self.assertEqual(2, playfield.available_balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)
        self._captured = 0

        # it should eject one
        self.assertTrue(coil5.pulse.called)
        coil5.pulse = MagicMock()
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch("s_ball_switch_target3",
                                                      0)

        # wait for confirm
        self.advance_time_and_run(11)
        self.assertEqual(1, playfield.balls)

        # it should eject the second
        self.assertTrue(coil5.pulse.called)
        coil5.pulse = MagicMock()
        self.advance_time_and_run(.1)
        self.machine.switch_controller.process_switch(
            "s_ball_switch_target3_2", 0)

        # wait for confirm
        self.advance_time_and_run(11)
        self.assertEqual(2, playfield.balls)
        self.assertEqual(0, self._captured)
        self.assertEqual(2, playfield.available_balls)

    def test_ball_request_when_device_is_full(self):
        coil1 = self.machine.coils['eject_coil1']
        coil2 = self.machine.coils['eject_coil2']
        coil3 = self.machine.coils['eject_coil3']
        trough = self.machine.ball_devices['test_trough']
        launcher = self.machine.ball_devices['test_launcher']
        target = self.machine.ball_devices['test_target1']
        playfield = self.machine.ball_devices['playfield']

        self.machine.events.add_handler('balldevice_captured_from_playfield',
                                        self._captured_from_pf)
        self.machine.events.add_handler('balldevice_ball_missing',
                                        self._missing_ball)
        self.machine.events.add_handler('collecting_balls_complete',
                                        self._collecting_balls_complete_handler)

        self.assertEqual(None, self.machine.game)

        self._enter = 0
        self._captured = 0
        self._missing = 0
        self._collecting_balls_complete = 0

        # add initial balls to trough
        self.machine.switch_controller.process_switch("s_ball_switch1", 1)
        self.machine.switch_controller.process_switch("s_ball_switch2", 1)
        self.advance_time_and_run(1)
        self.assertEqual(2, self._captured)
        self._captured = 0
        self.assertEqual(0, playfield.balls)
        self.assertEqual(2, self.machine.ball_controller.num_balls_known)

        # it should keep the ball
        coil1.pulse = MagicMock()
        coil2.pulse = MagicMock()
        coil3.pulse = MagicMock()
        self.assertEqual(2, trough.balls)
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # launcher requests a ball
        launcher.request_ball()
        self.advance_time_and_run(1)

        # trough ejects
        self.assertTrue(coil1.pulse.called)
        coil1.pulse = MagicMock()
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch1", 0)
        self.advance_time_and_run(1)
        self.assertEqual(1, trough.balls)

        # launcher receives and ejects
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)
        self.assertEqual(1, launcher.available_balls)

        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # launcher requests a second ball (but already has one)
        launcher.request_ball()
        self.advance_time_and_run(1)

        # launcher is full so nothing should happen
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called
        self.assertEqual(1, launcher.balls)
        self.assertEqual(2, launcher.available_balls)

        # target1 requests a ball
        target.request_ball()
        self.advance_time_and_run(1)

        # launcher shoots the ball
        assert not coil1.pulse.called
        self.assertTrue(coil2.pulse.called)
        coil2.pulse = MagicMock()
        assert not coil3.pulse.called

        # ball leaves
        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      0)
        self.advance_time_and_run(1)
        self.assertEqual(0, launcher.balls)

        # no eject of trough yet
        assert not coil1.pulse.called
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        # target receives
        self.machine.switch_controller.process_switch("s_ball_switch_target1",
                                                      1)
        self.advance_time_and_run(1)

        # trough ejects
        self.assertTrue(coil1.pulse.called)
        coil1.pulse = MagicMock()
        assert not coil2.pulse.called
        assert not coil3.pulse.called

        self.machine.switch_controller.process_switch("s_ball_switch2", 0)
        self.advance_time_and_run(1)
        self.assertEqual(0, trough.balls)

        self.machine.switch_controller.process_switch("s_ball_switch_launcher",
                                                      1)
        self.advance_time_and_run(1)
        self.assertEqual(1, launcher.balls)

        self.assertEqual("idle", trough._state)
        self.assertEqual("idle", launcher._state)
        self.assertEqual("idle", target._state)

    def test_entrance_switch_ignore_window(self):
        """Verify that an entrance switch doesn't hit during the ignore window."""
        self.mock_event("balldevice_captured_from_playfield")

        # First call should register the capture
        self.machine.switch_controller.process_switch("s_entrance", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_entrance", 0)
        self.advance_time_and_run(0.1)
        self.assertEventCalled("balldevice_captured_from_playfield", times=1)

        # Second call, 2 seconds later, should be ignored
        self.advance_time_and_run(2)
        self.machine.switch_controller.process_switch("s_entrance", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_entrance", 0)
        self.advance_time_and_run(0.1)
        self.assertEventCalled("balldevice_captured_from_playfield", times=1)

        # Third call, 2 seconds later, should capture
        self.advance_time_and_run(2)
        self.machine.switch_controller.process_switch("s_entrance", 1)
        self.advance_time_and_run(0.1)
        self.machine.switch_controller.process_switch("s_entrance", 0)
        self.advance_time_and_run(0.1)
        self.assertEventCalled("balldevice_captured_from_playfield", times=2)



"""Templates and placeholders."""
import ast
import string
import asyncio
import operator as op
import abc
from functools import lru_cache

import re
from typing import Tuple, List, Any

from mpf.core.utility_functions import Util

from mpf.core.mpf_controller import MpfController

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import


# supported operators
OPERATORS = {ast.Add: op.add, ast.Sub: op.sub, ast.Mult: op.mul,
             ast.Div: op.truediv, ast.Pow: op.pow, ast.BitXor: op.xor,
             ast.USub: op.neg, ast.Not: op.not_, ast.Mod: op.mod}

BOOL_OPERATORS = {ast.And: lambda a, b: a and b, ast.Or: lambda a, b: a or b}

COMPARISONS = {ast.Eq: op.eq, ast.Lt: op.lt, ast.Gt: op.gt, ast.LtE: op.le, ast.GtE: op.ge, ast.NotEq: op.ne}


class ConditionalEvent:

    """An conditional event."""

    __slots__ = ["name", "condition", "number"]

    def __init__(self, name, condition, number):
        """Initialise conditional event."""
        self.name = name
        self.condition = condition
        self.number = number

    def __repr__(self):
        """Return string representation."""
        return "{}-{}-{}".format(self.name, self.condition, self.number)


class TemplateEvalError(Exception):

    """An error occurred during a template evaluation."""

    def __init__(self, subscriptions):
        """Remember subscriptions."""
        super().__init__(subscriptions)
        self.subscriptions = subscriptions

    def __str__(self):
        """Return description."""
        return "<TemplateEvalError with subscriptions {}>".format(self.subscriptions)


class BaseTemplate(metaclass=abc.ABCMeta):

    """Base class for templates."""

    __slots__ = ["template", "placeholder_manager", "default_value", "text"]

    def __init__(self, template, text, placeholder_manger, default_value):
        """Initialise template."""
        self.text = str(text)
        self.template = template
        self.placeholder_manager = placeholder_manger
        self.default_value = default_value

    def evaluate(self, parameters, fail_on_missing_params=False):
        """Evaluate template and convert the result."""
        try:
            result = self.placeholder_manager.evaluate_template(self.template, parameters)
        except ValueError:
            if fail_on_missing_params:
                raise
            return self.default_value
        except TemplateEvalError:
            return self.default_value
        if result is None:
            return self.default_value
        return self.convert_result(result)

    def evaluate_or_none(self, parameters):
        """Evaluate template and convert the result or return None."""
        try:
            result = self.placeholder_manager.evaluate_template(self.template, parameters)
        except ValueError:
            return None
        if result is None:
            return None
        return self.convert_result(result)

    def evaluate_and_subscribe(self, parameters) -> Tuple[bool, asyncio.Future]:
        """Evaluate template and subscribe."""
        result, subscriptions = self.placeholder_manager.evaluate_and_subscribe_template(self.template, parameters,
                                                                                         self.text)
        if isinstance(result, TemplateEvalError) or result is None:
            result = self.default_value
        return self.convert_result(result), subscriptions

    @abc.abstractmethod
    def convert_result(self, value):
        """Convert the result of the template."""

    def __repr__(self):
        """Return string representation."""
        return "<Template {}>".format(self.text)


class BoolTemplate(BaseTemplate):

    """Bool template."""

    __slots__ = []  # type: List[str]

    def convert_result(self, value):
        """Convert the result to bool."""
        return bool(value)


class FloatTemplate(BaseTemplate):

    """Float template."""

    __slots__ = []  # type: List[str]

    def convert_result(self, value):
        """Convert the result to float."""
        return float(value)


class IntTemplate(BaseTemplate):

    """Float template."""

    __slots__ = []  # type: List[str]

    def convert_result(self, value):
        """Convert the result to int."""
        return int(value)


class StringTemplate(BaseTemplate):

    """String template."""

    __slots__ = []  # type: List[str]

    def convert_result(self, value):
        """Convert the result to str."""
        return str(value)


class RawTemplate(BaseTemplate):

    """Raw template."""

    __slots__ = []  # type: List[str]

    def convert_result(self, value):
        """Keep the value."""
        return value


class NativeTypeTemplate:

    """Native type template which encapsulates an int/float/bool."""

    __slots__ = ["value", "machine"]

    def __init__(self, value, machine):
        """Set value."""
        self.value = value
        self.machine = machine

    def evaluate(self, parameters, fail_on_missing_params=False):
        """Return value."""
        del parameters
        del fail_on_missing_params
        return self.value

    def evaluate_or_none(self, parameters):
        """Return value."""
        del parameters
        return self.value

    def evaluate_and_subscribe(self, parameters) -> Tuple[int, asyncio.Future]:
        """Evaluate and subscribe template."""
        del parameters
        future = asyncio.Future()   # type: asyncio.Future
        return self.value, future

    def __eq__(self, other):
        """Templates are equal if values are equal."""
        return other.value == self.value

    def __repr__(self):
        """Return String."""
        return "<NativeTemplate {}>".format(self.value)


class MpfFormatter(string.Formatter):

    """String formater which replaces placeholders."""

    __slots__ = ["machine", "parameters", "subscriptions", "subscribe"]

    def __init__(self, machine, parameters, subscribe):
        """Initialise formatter."""
        self.machine = machine
        self.parameters = parameters
        self.subscriptions = []
        self.subscribe = subscribe

    def get_value(self, key, args, kwargs):
        """Return value of placeholder."""
        placeholder = self.machine.placeholder_manager.build_raw_template(key)
        if self.subscribe:
            value, future = placeholder.evaluate_and_subscribe(self.parameters)
            if future:
                self.subscriptions.append(future)
            return value

        return placeholder.evaluate(self.parameters)

    def get_field(self, field_name, args, kwargs):
        """Return value of field."""
        obj = self.get_value(field_name, args, kwargs)
        return obj, field_name

    def format_field(self, value, format_spec):
        """Format field."""
        # don't crash on None for int. the format type is always the last element in a format spec
        if value is None and format_spec[-1:] == "d":
            value = 0
        return super().format_field(value, format_spec)


class TextTemplate:

    """Text placeholder."""

    __slots__ = ["machine", "text", "_change_callback"]

    def __init__(self, machine: "MachineController", text: str) -> None:
        """Initialise placeholder."""
        self.machine = machine
        self.text = str(text)
        self._change_callback = None

    def evaluate(self, parameters) -> str:
        """Evaluate placeholder to string."""
        f = MpfFormatter(self.machine, parameters, False)
        return f.format(self.text)

    def evaluate_and_subscribe(self, parameters) -> Tuple[str, asyncio.Future]:
        """Evaluate placeholder to string and subscribe to changes."""
        f = MpfFormatter(self.machine, parameters, True)
        value = f.format(self.text)
        subscriptions = f.subscriptions
        if not subscriptions:
            future = asyncio.Future()   # type: asyncio.Future
        elif len(subscriptions) == 1:
            future = subscriptions[0]
        else:
            future = Util.any(subscriptions)
        future = asyncio.ensure_future(future)
        return value, future


class BasePlaceholder:

    """Base class for placeholder variables."""

    __slots__ = []  # type: List[str]

    # pylint: disable-msg=no-self-use
    def subscribe(self):
        """Subscribe to placeholder."""
        raise AssertionError("Not possible to subscribe this.")

    # pylint: disable-msg=no-self-use
    def subscribe_attribute(self, item):
        """Subscribe to attribute."""
        raise AssertionError("Not possible to subscribe to attribute {}.".format(item))

    # pylint: disable-msg=no-self-use
    def subscribe_item(self, item):
        """Subscribe to item."""
        raise AssertionError("Not possible to subscribe to item {}.".format(item))


class DevicePlaceholder:

    """Wrap a monitorable device."""

    __slots__ = ["_device", "_attribute", "_machine"]

    def __init__(self, device, attribute, machine):
        """Initialise placeholder."""
        self._device = device
        self._attribute = attribute
        self._machine = machine

    @staticmethod
    def subscribe():
        """Subscribe to object changes."""
        return asyncio.Future()

    def subscribe_attribute(self, item):
        """Subscribe to device changes."""
        return self._device.subscribe_attribute(item, self._machine)

    def __getitem__(self, item):
        """Array access."""
        return self.__getattr__(item)

    def __getattr__(self, item):
        """Attribute access."""
        return self._device.get_placeholder_value(item)


class DeviceClassPlaceholder:

    """Wrap a monitorable device class."""

    __slots__ = ["_devices", "_device_name", "_machine"]

    def __init__(self, devices, device_name, machine):
        """Initialise placeholder."""
        self._devices = devices
        self._device_name = device_name
        self._machine = machine

    @staticmethod
    def subscribe():
        """Subscribe to object changes."""
        return asyncio.Future()

    @staticmethod
    def subscribe_attribute(item):
        """Subscribe to device changes."""
        del item
        return asyncio.Future()

    def __getitem__(self, item):
        """Array access."""
        return self.__getattr__(item)

    def __getattr__(self, item):
        """Attribute access."""
        device = self._devices.get(item)
        if not device:
            raise AssertionError("Device {} of type {} does not exist.".format(item, self._device_name))

        return DevicePlaceholder(device, item, self._machine)


class DevicesPlaceholder:

    """Device monitor placeholder."""

    __slots__ = ["_machine"]

    def __init__(self, machine):
        """Initialise placeholder."""
        self._machine = machine

    def __getitem__(self, item):
        """Array access."""
        return self.__getattr__(item)

    @staticmethod
    def subscribe():
        """Subscribe to object changes."""
        return asyncio.Future()

    @staticmethod
    def subscribe_attribute(item):
        """Subscribe to device changes."""
        del item
        return asyncio.Future()

    def __getattr__(self, item):
        """Attribute access."""
        device = self._machine.device_manager.get_monitorable_devices().get(item)
        if not device:
            raise AssertionError("Device Collection {} not usable in placeholders.".format(item))
        return DeviceClassPlaceholder(device, item, self._machine)


class ModeClassPlaceholder:

    """Wrap a mode."""

    __slots__ = ["_mode"]

    def __init__(self, mode):
        """Initialise placeholder."""
        self._mode = mode

    def __getitem__(self, item):
        """Array access."""
        return self.__getattr__(item)

    def __getattr__(self, item):
        """Attribute access."""
        this_item = getattr(self._mode, item)
        return this_item


class ModePlaceholder:

    """Mode placeholder."""

    __slots__ = ["_machine"]

    def __init__(self, machine):
        """Initialise placeholder."""
        self._machine = machine

    def __getitem__(self, item):
        """Array access."""
        return self.__getattr__(item)

    def __getattr__(self, item):
        """Attribute access."""
        if item not in self._machine.modes:
            raise ValueError("{} is not a valid mode name".format(item))

        return ModeClassPlaceholder(self._machine.modes[item])


class PlayerPlaceholder(BasePlaceholder):

    """Wraps the player."""

    __slots__ = ["_machine", "_number"]

    def __init__(self, machine, number=None):
        """Initialise placeholder."""
        self._machine = machine     # type: MachineController
        self._number = number

    def subscribe(self):
        """Subscribe to player changes."""
        return self._machine.events.wait_for_any_event(["player_turn_ended", "player_turn_started"])

    def subscribe_attribute(self, item):
        """Subscribe player variable changes."""
        return self._machine.events.wait_for_event('player_{}'.format(item))

    def __getitem__(self, item):
        """Array access."""
        if self._machine.game and self._machine.game.player:
            if self._number is not None:
                if len(self._machine.game.player_list) <= self._number:
                    raise ValueError("Player not in game")
                return self._machine.game.player_list[self._number][item]

            return self._machine.game.player[item]

        raise ValueError("Not in a game")

    def __getattr__(self, item):
        """Attribute access."""
        if self._machine.game and self._machine.game.player:
            if self._number is not None:
                if len(self._machine.game.player_list) <= self._number:
                    raise ValueError("Player not in game")
                return getattr(self._machine.game.player_list[self._number], item)

            return getattr(self._machine.game.player, item)

        raise ValueError("Not in a game")


class PlayersPlaceholder(BasePlaceholder):

    """Wraps the player list."""

    __slots__ = ["_machine"]

    def __init__(self, machine):
        """Initialise placeholder."""
        self._machine = machine     # type: MachineController

    def subscribe(self):
        """Subscribe to player list changes."""
        return self._machine.events.wait_for_any_event(["player_added", "game_ended"])

    def subscribe_attribute(self, item):
        """Subscribe player variable changes."""
        return self._machine.events.wait_for_event('player_{}'.format(item))

    def __getitem__(self, item):
        """Array access."""
        return PlayerPlaceholder(self._machine, item)

    def __getattr__(self, item):
        """Attribute access."""
        return PlayerPlaceholder(self._machine, item)


class MachinePlaceholder(BasePlaceholder):

    """Wraps the machine."""

    __slots__ = ["_machine"]

    def __init__(self, machine):
        """Initialise placeholder."""
        self._machine = machine     # type: MachineController

    def subscribe(self):
        """Subscribe to machine changes.

        Will never return.
        """
        return asyncio.Future()

    def subscribe_attribute(self, item):
        """Subscribe to machine variable."""
        return self._machine.events.wait_for_event('machine_var_{}'.format(item))

    def __getitem__(self, item):
        """Array access."""
        return self._machine.variables.get_machine_var(item)

    def __getattr__(self, item):
        """Attribute access."""
        return self._machine.variables.get_machine_var(item)


class SettingsPlaceholder(BasePlaceholder):

    """Wraps settings."""

    __slots__ = ["_machine"]

    def __init__(self, machine):
        """Initialise placeholder."""
        self._machine = machine  # type: MachineController

    def subscribe(self):
        """Subscribe to settings controller changes.

        Will never return.
        """
        return asyncio.Future()

    def subscribe_attribute(self, item):
        """Subscribe to machine variable for this setting."""
        return self._machine.events.wait_for_event(
            'machine_var_{}'.format(self._machine.settings.get_setting_machine_var(item)))

    def __getattr__(self, item):
        """Attribute access."""
        return self._machine.settings.get_setting_value(item)


class BasePlaceholderManager(MpfController):

    """Manages templates and placeholders for MPF and MC."""

    # needed here so the auto-detection of child classes works
    module_name = 'PlaceholderManager'
    config_name = 'placeholder_manager'

    __slots__ = ["_eval_methods"]

    def __init__(self, machine):
        """Initialise."""
        super().__init__(machine)
        self._eval_methods = {
            ast.Num: self._eval_num,
            ast.Str: self._eval_str,
            ast.NameConstant: self._eval_constant,
            ast.BinOp: self._eval_bin_op,
            ast.UnaryOp: self._eval_unary_op,
            ast.Compare: self._eval_compare,
            ast.BoolOp: self._eval_bool_op,
            ast.Attribute: self._eval_attribute,
            ast.Subscript: self._eval_subscript,
            ast.Name: self._eval_name,
            ast.IfExp: self._eval_if
        }
        if hasattr(ast, "Constant"):
            self._eval_methods[ast.Constant] = self._eval_constant

    @staticmethod
    def _parse_template(template_str):
        try:
            return ast.parse(template_str, mode='eval').body
        except SyntaxError:
            raise AssertionError('Failed to parse template "{}"'.format(template_str))

    @staticmethod
    def _eval_num(node, variables, subscribe):
        del variables
        del subscribe
        return node.n, []

    @staticmethod
    def _eval_str(node, variables, subscribe):
        del variables
        del subscribe
        return node.s, []

    @staticmethod
    def _eval_constant(node, variables, subscribe):
        del variables
        del subscribe
        return node.value, []

    def _eval_if(self, node, variables, subscribe):
        value, subscription = self._eval(node.test, variables, subscribe)
        if value:
            ret_value, ret_subscription = self._eval(node.body, variables, subscribe)
            return ret_value, subscription + ret_subscription

        ret_value, ret_subscription = self._eval(node.orelse, variables, subscribe)
        return ret_value, subscription + ret_subscription

    def _eval_bin_op(self, node, variables, subscribe):
        left_value, left_subscription = self._eval(node.left, variables, subscribe)
        right_value, right_subscription = self._eval(node.right, variables, subscribe)
        try:
            ret_value = OPERATORS[type(node.op)](left_value, right_value)
        except TypeError:
            raise TemplateEvalError(left_subscription + right_subscription)
        return ret_value, left_subscription + right_subscription

    def _eval_unary_op(self, node, variables, subscribe):
        value, subscription = self._eval(node.operand, variables, subscribe)
        return OPERATORS[type(node.op)](value), subscription

    def _eval_compare(self, node, variables, subscribe):
        if len(node.ops) > 1:
            raise AssertionError("Only single comparisons are supported.")
        left_value, left_subscription = self._eval(node.left, variables, subscribe)
        right_value, right_subscription = self._eval(node.comparators[0], variables, subscribe)
        try:
            return COMPARISONS[type(node.ops[0])](left_value, right_value), left_subscription + right_subscription
        except TypeError:
            raise TemplateEvalError(left_subscription + right_subscription)

    def _eval_bool_op(self, node, variables, subscribe):
        result, subscription = self._eval(node.values[0], variables, subscribe)
        for i in range(1, len(node.values)):
            value, new_subscription = self._eval(node.values[i], variables, subscribe)
            subscription += new_subscription
            try:
                result = BOOL_OPERATORS[type(node.op)](result, value)
            except TypeError:
                raise TemplateEvalError(subscription)
        return result, subscription

    def _eval_attribute(self, node, variables, subscribe):
        slice_value, subscription = self._eval(node.value, variables, subscribe)
        if isinstance(slice_value, dict) and node.attr in slice_value:
            ret_value = slice_value[node.attr]
        else:
            try:
                ret_value = getattr(slice_value, node.attr)
            except (ValueError, AttributeError):
                if subscribe:   # pylint: disable-msg=no-else-raise
                    raise TemplateEvalError(subscription + [slice_value.subscribe_attribute(node.attr)])
                else:
                    raise
        if subscribe:
            return ret_value, subscription + [slice_value.subscribe_attribute(node.attr)]

        return ret_value, subscription + []

    def _eval_subscript(self, node, variables, subscribe):
        value, subscription = self._eval(node.value, variables, subscribe)
        if isinstance(node.slice, ast.Constant):
            return value[node.slice.value], subscription
        if isinstance(node.slice, ast.Index):
            slice_value, slice_subscript = self._eval(node.slice.value, variables, subscribe)
            try:
                return value[slice_value], subscription + slice_subscript
            except ValueError:
                raise TemplateEvalError(subscription + slice_subscript)
        if isinstance(node.slice, ast.Slice):
            lower, lower_subscription = self._eval(node.slice.lower, variables, subscribe)
            upper, upper_subscription = self._eval(node.slice.upper, variables, subscribe)
            step, step_subscription = self._eval(node.slice.step, variables, subscribe)
            return value[lower:upper:step], subscription + lower_subscription + upper_subscription + step_subscription

        raise TypeError(type(node.slice))

    def _eval_name(self, node, variables, subscribe):
        if node.id in ("true", "false"):
            self.raise_config_error("Placeholder use Python syntax. Use True "
                                    "and False instead of true and false.", 1,
                                    context=node.id)

        var = self.get_global_parameters(node.id)
        if var:
            if subscribe:
                return var, [var.subscribe()]

            return var, []
        if node.id in variables:
            return variables[node.id], []

        raise ValueError("Missing variable {}".format(node.id))

    def _eval(self, node, variables, subscribe) -> Tuple[Any, List]:
        if node is None:
            return None, []

        if type(node) in self._eval_methods:  # pylint: disable-msg=unidiomatic-typecheck
            return self._eval_methods[type(node)](node, variables, subscribe)

        raise TypeError(type(node))

    def build_float_template(self, template_str, default_value=0.0) -> FloatTemplate:
        """Build a float template from a string."""
        # try to convert to int
        try:
            value = float(template_str)
        except ValueError:
            pass
        else:
            return NativeTypeTemplate(value, self.machine)  # type: ignore

        return FloatTemplate(self._parse_template(template_str), template_str, self, default_value)

    def build_int_template(self, template_str, default_value=0) -> IntTemplate:
        """Build a int template from a string."""
        # try to convert to int
        try:
            value = int(template_str)
        except ValueError:
            pass
        else:
            return NativeTypeTemplate(value, self.machine)  # type: ignore

        return IntTemplate(self._parse_template(template_str), template_str, self, default_value)

    def build_bool_template(self, template_str, default_value=False) -> BoolTemplate:
        """Build a bool template from a string."""
        if isinstance(template_str, bool):
            return NativeTypeTemplate(template_str, self.machine)   # type: ignore
        return BoolTemplate(self._parse_template(template_str), template_str, self, default_value)

    def build_string_template(self, template_str, default_value=""):
        """Build a string template from a string."""
        return StringTemplate(self._parse_template(template_str), template_str, self, default_value)

    def build_text_template(self, template_str):
        """Build a full featured text template."""
        return TextTemplate(self.machine, template_str)

    def build_quoted_string_template(self, template_str, default_value=""):
        """Build a string template from a string if enclosed in brackets."""
        if isinstance(template_str, str) and (template_str[0:1] != "(" or template_str[-1:] != ")"):
            return NativeTypeTemplate(template_str, self.machine)
        return StringTemplate(self._parse_template(template_str), template_str, self, default_value)

    def build_raw_template(self, template_str, default_value=None) -> RawTemplate:
        """Build a raw template from a string."""
        return RawTemplate(self._parse_template(template_str), template_str, self, default_value)

    def get_global_parameters(self, name):
        """Return global params."""
        raise NotImplementedError()

    def evaluate_template(self, template, parameters):
        """Evaluate template."""
        return self._eval(template, parameters, False)[0]

    def evaluate_and_subscribe_template(self, template, parameters, text=None):
        """Evaluate and subscribe template."""
        try:
            value, subscriptions = self._eval(template, parameters, True)
        except TemplateEvalError as e:
            value = e
            subscriptions = e.subscriptions
        except ValueError as e:
            raise AssertionError("Failed to evaluate and subscribe template {} with parameters {}. "
                                 "See error above.".format(text, parameters)) from e

        if not subscriptions:
            future = asyncio.Future()
        elif len(subscriptions) == 1:
            future = subscriptions[0]
        else:
            future = Util.any(subscriptions)
        future = asyncio.ensure_future(future)
        return value, future

    @lru_cache(typed=True)
    def parse_conditional_template(self, template, default_number=None):
        """Parse a template for condition and number and return a dict."""
        # The following regex will make a dict for event name, condition, and number
        # e.g. some_event_name_string{variable.condition==True}|num
        #      ^ string at start     ^ condition in braces     ^ pipe- or colon-delimited value
        match = re.search(r"^(?P<name>[^{}:| ]+)({(?P<condition>.+)})?([|:](?P<number>.+))?$", template)
        if not match:
            raise AssertionError("Invalid template string {}".format(template))

        match_dict = match.groupdict()

        # Create a Template object for the condition
        if match_dict['condition'] is not None:
            match_dict['condition'] = self.build_bool_template(match_dict['condition'])

        if default_number is not None:
            # Fill in the default number if the template has none
            if match_dict['number'] is None:
                match_dict['number'] = default_number
            else:
                # Type-conform the template number to the default_number type
                try:
                    match_dict['number'] = type(default_number)(match_dict['number'])
                # Gracefully fall back if the number can't be parsed
                except ValueError:
                    self.warning_log("Condition '{}' has invalid number value '{}'".format(
                                     template, match_dict['number']))
                    match_dict['number'] = default_number
        return ConditionalEvent(match_dict["name"], match_dict["condition"], match_dict["number"])


class PlaceholderManager(BasePlaceholderManager):

    """Manages templates and placeholders for MPF."""

    __slots__ = []  # type: List[str]

    # pylint: disable-msg=too-many-return-statements
    def get_global_parameters(self, name):
        """Return global params."""
        if name == "settings":
            return SettingsPlaceholder(self.machine)
        if name == "machine":
            return MachinePlaceholder(self.machine)
        if name == "device":
            return DevicesPlaceholder(self.machine)
        if name == "mode":
            return ModePlaceholder(self.machine)
        if name == "current_player":
            return PlayerPlaceholder(self.machine)
        if name == "players":
            return PlayersPlaceholder(self.machine)
        if self.machine.game:
            if name == "game":
                return self.machine.game

        return False



from functools import partial
from queue import Queue
from unittest.mock import MagicMock, patch, call

from mpf.tests.MpfBcpTestCase import MockBcpClient
from mpf.tests.MpfTestCase import MpfTestCase


class TestBcpClient(MockBcpClient):
    def __init__(self, machine, name, bcp):
        super().__init__(machine, name, bcp)
        self.queue = Queue()
        self.exit_on_close = False

    def send(self, bcp_command, kwargs):
        if bcp_command == "reset":
            self.receive_queue.put_nowait(("reset_complete", {}))
        self.queue.put((bcp_command, kwargs))

    def receive(self, bcp_command, callback=None, rawbytes=None, **kwargs):
        if rawbytes:
            kwargs['rawbytes'] = rawbytes
        self.receive_queue.put_nowait((bcp_command, kwargs))

        if callback:
            callback()

class MockMcClock:

    def __init__(self, clock):
        self._clock = clock

    def __getattr__(self, item):
        return getattr(self._clock, item)

    def schedule_once(self, callback, timeout=0):
        self._clock.schedule_once(partial(callback, dt=None), timeout)

    def schedule_interval(self, callback, timeout):
        self._clock.schedule_interval(partial(callback, dt=None), timeout)


class TestBcp(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        # remove config patch which disables bcp
        self.machine_config_patches['bcp'] = \
            {"connections": {"local_display": {"type": "mpf.tests.test_BcpMc.TestBcpClient"}}}
        self.machine_config_patches['bcp']['servers'] = []

    def get_use_bcp(self):
        return True

    def setUp(self):
        super().setUp()

        client = self.machine.bcp.transport.get_named_client("local_display")
        self.kivy = MagicMock()
        self.kivy.clock.Clock = MockMcClock(self.machine.clock)
        modules = {
            'kivy': self.kivy,
            'kivy.clock': self.kivy.clock,
            'kivy.logger': self.kivy.logger,
        }
        self.module_patcher = patch.dict('sys.modules', modules)
        self.module_patcher.start()

        try:
            from mpfmc.core import bcp_processor
        except ImportError:
            super().tearDown()
            self.skipTest("Cannot import mpfmc.core.bcp_processor")
            return

        self.mc = MagicMock()
        bcp_processor.BcpProcessor._start_socket_thread = MagicMock()
        bcp_mc = bcp_processor.BcpProcessor(self.mc)
        bcp_mc.enabled = True
        bcp_mc.send = client.receive
        bcp_mc._client_connected()
        self.machine_run()
        self.mc.events.post = MagicMock()
        while not client.queue.empty():
            bcp_mc.receive_queue.put(client.queue.get())
        client.queue = bcp_mc.receive_queue
        self.machine_run()

    def tearDown(self):
        self.machine_run()
        self.module_patcher.stop()
        super().tearDown()

    def test_bcp_mpf_and_mpf_mc(self):
        client = self.machine.bcp.transport.get_named_client("local_display")
        self.machine.bcp.interface.add_registered_trigger_event_for_client(client, 'ball_started')
        self.machine.bcp.interface.add_registered_trigger_event_for_client(client, 'ball_ended')
        self.machine.events.post('ball_started', ball=17, player=23)

        self.machine_run()

        self.mc.events.post.assert_has_calls([call("ball_started", ball=17, player=23), ])
        self.mc.events.post.reset_mock()

        self.machine.events.post('ball_ended')
        self.machine_run()
        self.mc.events.post.assert_has_calls([
            call("ball_ended")
        ])



"""Test the bonus mode."""
import os
import time
from unittest.mock import mock_open, patch

from mpf.file_interfaces.yaml_interface import YamlInterface
from mpf.core.data_manager import DataManager
from mpf.tests.MpfTestCase import MpfTestCase


class TestDataManager(MpfTestCase):

    def get_config_file(self):
        return "config.yaml"

    def get_machine_path(self):
        return 'tests/machine_files/data_manager/'

    def setUp(self):
        super().setUp()
        YamlInterface.cache = False

    def tearDown(self):
        YamlInterface.cache = True
        super().tearDown()

    def test_save_and_load(self):
        open_mock = mock_open(read_data="")
        with patch('mpf.file_interfaces.yaml_interface.open', open_mock, create=True):
            with patch('mpf.core.file_manager.os.path.isfile') as isfile_mock:
                isfile_mock.return_value = True
                manager = DataManager(self.machine, "machine_vars", min_wait_secs=0)
                self.assertTrue(isfile_mock.called)
                self.assertTrue(open_mock.called)

        self.assertNotIn("hallo", manager.get_data())

        open_mock = mock_open(read_data="")
        with patch('mpf.file_interfaces.yaml_interface.open', open_mock, create=True):
            with patch('mpf.core.data_manager.os.replace') as move_mock:
                manager.save_all({"hallo": "world"})
                while not move_mock.called:
                    time.sleep(.00001)
                open_mock().write.assert_called_once_with('hallo: world\n')
                self.assertTrue(move_mock.called)

        open_mock = mock_open(read_data='hallo: world\n')
        with patch('mpf.file_interfaces.yaml_interface.open', open_mock, create=True):
            with patch('mpf.core.file_manager.os.path.isfile') as isfile_mock:
                isfile_mock.return_value = True
                manager2 = DataManager(self.machine, "machine_vars", min_wait_secs=0)
                self.assertTrue(isfile_mock.called)
                self.assertTrue(open_mock.called)

        self.assertEqual("world", manager2.get_data()["hallo"])
        self.assertEqual({}, manager.get_data("hallo"))

    def test_get_data(self):
        open_mock = mock_open(read_data='hallo:\n  test: world\n')
        with patch('mpf.file_interfaces.yaml_interface.open', open_mock, create=True):
            with patch('mpf.core.file_manager.os.path.isfile') as isfile_mock:
                manager = DataManager(self.machine, "machine_vars", min_wait_secs=0)
                self.assertTrue(isfile_mock.called)
                self.assertTrue(open_mock.called)

        self.assertEqual({"test": "world"}, manager.get_data("hallo"))
        self.assertEqual({}, manager.get_data("invalid"))

    def test_paths_and_disable(self):
        open_mock = mock_open(read_data='hallo:\n  test: world\n')
        with patch('mpf.file_interfaces.yaml_interface.open', open_mock, create=True):
            with patch('mpf.core.file_manager.os.path.isfile') as isfile_mock:
                with patch('mpf.core.file_manager.os.mkdir') as mkdir_mock:
                    manager = DataManager(self.machine, "absolute_test", min_wait_secs=0)
                    isfile_mock.assert_called_with('/data/test_dir/test_file.yaml')
                    mkdir_mock.assert_called_with('/data/test_dir', 511)
                    open_mock.assert_called_once_with('/data/test_dir/test_file.yaml', encoding='utf8')

        self.assertEqual({"test": "world"}, manager.get_data("hallo"))
        self.assertEqual({}, manager.get_data("invalid"))

        open_mock = mock_open(read_data='hallo:\n  test: world\n')
        with patch('mpf.file_interfaces.yaml_interface.open', open_mock, create=True):
            with patch('mpf.core.file_manager.os.path.isfile') as isfile_mock:
                with patch('mpf.core.file_manager.os.mkdir') as mkdir_mock:
                    manager = DataManager(self.machine, "relative_test", min_wait_secs=0)
                    path = os.path.join(os.path.abspath(self.machine.machine_path), 'subdir/subdir2')
                    file_path = os.path.join(os.path.abspath(self.machine.machine_path), 'subdir/subdir2/test.yaml')
                    isfile_mock.assert_called_with(file_path)
                    mkdir_mock.assert_called_with(path, 511)
                    open_mock.assert_called_once_with(file_path, encoding='utf8')

        self.assertEqual({"test": "world"}, manager.get_data("hallo"))
        self.assertEqual({}, manager.get_data("invalid"))
        open_mock = mock_open(read_data='hallo:\n  test: world\n')
        with patch('mpf.file_interfaces.yaml_interface.open', open_mock, create=True):
            with patch('mpf.core.file_manager.os.path.isfile') as isfile_mock:
                with patch('mpf.core.file_manager.os.mkdir') as mkdir_mock:
                    manager = DataManager(self.machine, "disabled_test", min_wait_secs=0)
                    self.assertFalse(isfile_mock.called)
                    self.assertFalse(mkdir_mock.called)
                    self.assertFalse(open_mock.called)

        self.assertEqual({}, manager.get_data("hallo"))
        self.assertEqual({}, manager.get_data("invalid"))



"""Config player for sounds on an external sound card."""
from typing import List

from mpf.config_players.device_config_player import DeviceConfigPlayer

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.hardware_sound_system import HardwareSoundSystem   # pylint: disable-msg=cyclic-import,unused-import; # noqa


class HardwareSoundPlayer(DeviceConfigPlayer):

    """Plays sounds on an external sound card."""

    config_file_section = 'hardware_sound_player'
    show_section = 'hardware_sounds'

    __slots__ = []  # type: List[str]

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Play sound on external card."""
        del kwargs
        del context
        del calling_context

        for item, s in settings.items():
            sound_system = s['sound_system']        # type: HardwareSoundSystem
            if "value" in s and s["value"]:
                item = s["value"]

            if s['action'] == "stop":
                sound_system.stop_all_sounds()
            elif s['action'] == "play":
                sound_system.play(item, s["track"])
            elif s['action'] == "play_file":
                sound_system.play_file(item, s.get("platform_options", {}), s["track"])
            elif s['action'] == "text_to_speech":
                sound_system.text_to_speech(item, s.get("platform_options", {}), s["track"])
            elif s['action'] == "set_volume":
                sound_system.set_volume(float(item), s["track"])
            elif s['action'] == "increase_volume":
                sound_system.increase_volume(float(item), s["track"])
            elif s['action'] == "decrease_volume":
                sound_system.decrease_volume(float(item), s["track"])
            else:
                raise AssertionError("Invalid action {}".format(s['action']))

    def get_express_config(self, value):
        """Parse express config."""
        return dict(action=value)

    def get_string_config(self, string):
        """Parse string config."""
        if string == "stop":
            return {string: dict(action="stop")}

        return super().get_string_config(string)



import unittest
from unittest.mock import MagicMock

from mpf.core.bcp.bcp_socket_client import decode_command_string, encode_command_string
from mpf.tests.MpfTestCase import MpfTestCase
from mpf.tests.loop import MockQueueSocket


class TestBcpSocketClientEncoding(unittest.TestCase):

    def test_roundtrip(self):
        test_kwargs = {"registered_handlers": [{"calling_context": "logicblock_baby_bonus_hit{count%10==5}"}]}
        encoded_cmd = encode_command_string("test", **test_kwargs)
        decoded_cmd, decoded_cmd_kwargs = decode_command_string(encoded_cmd)
        self.assertEqual(decoded_cmd_kwargs, test_kwargs)

    def test_decode_command_string(self):
        # test strings
        string_in = 'play?some_value=7&another_value=2'
        command = 'play'
        kwargs = dict(some_value='7', another_value='2')

        actual_command, actual_kwargs = decode_command_string(string_in)

        self.assertEqual(command, actual_command)
        self.assertEqual(kwargs, actual_kwargs)

        # test ints, floats, bools, and none
        string_in = 'play?some_int=int:7&some_float=float:2.0&some_none' \
                    '=NoneType:&some_true=bool:True&some_false=bool:False'
        command = 'play'
        kwargs = dict(some_int=7, some_float=2.0, some_none=None,
                      some_true=True, some_false=False)

        actual_command, actual_kwargs = decode_command_string(string_in)

        self.assertEqual(command, actual_command)
        self.assertEqual(kwargs, actual_kwargs)

        string_in = "machine_variable?name=player1_score&prev_value=int:132990&value=&change=bool:True"
        actual_command, actual_kwargs = decode_command_string(string_in)
        self.assertEqual("machine_variable", actual_command)
        self.assertEqual({'prev_value': 132990, 'value': '', 'name': 'player1_score', 'change': True}, actual_kwargs)

    def test_encode_command_string(self):
        # test strings
        command = 'play'
        kwargs = dict(some_value='7', another_value='2')
        expected_strings = ('play?some_value=7&another_value=2',
                            'play?another_value=2&some_value=7')

        actual_string = encode_command_string(command, **kwargs)

        self.assertIn(actual_string, expected_strings)

        # test ints, floats, bools, and none
        command = 'play'
        kwargs = dict(some_int=7, some_float=2.0, some_none=None,
                      some_true=True, some_false=False)
        expected_snippets = ('play?',
                             'some_int=int:7',
                             'some_float=float:2.0',
                             'some_none=NoneType:',
                             'some_true=bool:True',
                             'some_false=bool:False')

        actual_string = encode_command_string(command, **kwargs)

        for snippet in expected_snippets:
            self.assertIn(snippet, actual_string)

    def test_json_encoding_decoding(self):
        # test with dicts
        command = 'play'
        kwargs = dict()
        kwargs['dict1'] = dict(key1='value1', key2='value2')
        kwargs['dict2'] = dict(key3='value3', key4='value4')

        encoded_string = encode_command_string(command, **kwargs)
        decoded_command, decoded_dict = decode_command_string(encoded_string)

        self.assertEqual('play', decoded_command)
        self.assertIn('dict1', decoded_dict)
        self.assertIn('dict2', decoded_dict)
        self.assertEqual(decoded_dict['dict1']['key1'], 'value1')
        self.assertEqual(decoded_dict['dict1']['key2'], 'value2')
        self.assertEqual(decoded_dict['dict2']['key3'], 'value3')
        self.assertEqual(decoded_dict['dict2']['key4'], 'value4')

        # test with list
        command = 'play'
        kwargs = dict()
        kwargs['dict1'] = dict(key1='value1', key2='value2')
        kwargs['dict2'] = list()
        kwargs['dict2'].append(dict(key3='value3', key4='value4'))
        kwargs['dict2'].append(dict(key3='value5', key4='value6'))

        encoded_string = encode_command_string(command, **kwargs)
        decoded_command, decoded_dict = decode_command_string(encoded_string)

        self.assertEqual('play', decoded_command)
        self.assertIn('dict1', decoded_dict)
        self.assertIn('dict2', decoded_dict)
        self.assertEqual(decoded_dict['dict1']['key1'], 'value1')
        self.assertEqual(decoded_dict['dict1']['key2'], 'value2')
        self.assertEqual(decoded_dict['dict2'][0],
                         dict(key3='value3', key4='value4'))
        self.assertEqual(decoded_dict['dict2'][1],
                         dict(key3='value5', key4='value6'))

    def test_json_encoding_decoding_with_hashtags(self):
        # see: https://github.com/missionpinball/mpf/issues/1495
        command = 'play'
        kwargs = dict()
        kwargs['dict1'] = dict(key1='value1 #', key2='value2 #')

        encoded_string = encode_command_string(command, **kwargs)
        decoded_command, decoded_dict = decode_command_string(encoded_string)

        self.assertEqual('play', decoded_command)
        self.assertIn('dict1', decoded_dict)
        self.assertEqual(decoded_dict['dict1']['key1'], 'value1 #')
        self.assertEqual(decoded_dict['dict1']['key2'], 'value2 #')


class MockBcpQueueSocket(MockQueueSocket):

    """Mock Queue Socket for BCP which emulates reset."""

    def send(self, data):
        if data == b'reset\n':
            self.recv_queue.append(b'reset_complete\n')
            return len(data)
        return super().send(data)


class TestBcpSocketClient(MpfTestCase):

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)

        # use normal client
        self.machine_config_patches['bcp'] = {}
        self.machine_config_patches['bcp']['servers'] = []
        self.receive_mock = None

    def get_use_bcp(self):
        return True

    def setUp(self):
        super().setUp()
        self._bcp_client = self.machine.bcp.transport.get_named_client("local_display")

    def _mock_loop(self):
        self.client_socket = MockBcpQueueSocket(self.loop)
        self.clock.mock_socket("localhost", 5050, self.client_socket)

    async def receive_func(self, *args, **kwargs):
        self.receive_mock(*args, **kwargs)

    def testReceiveMessages(self):
        # test message with bytes
        self.receive_mock = MagicMock()
        self.machine.bcp.interface.register_command_callback("receive_bytes", self.receive_func)
        self.client_socket.recv_queue.append(b'receive_bytes?name=default&bytes=4096\n')
        data = b'0' * 4096
        self.client_socket.recv_queue.append(data)
        self.advance_time_and_run()
        self.receive_mock.assert_called_once_with(name="default", client=self._bcp_client, rawbytes=data)
        self.receive_mock.reset_mock()

        # data and cmd in two packets
        self.client_socket.recv_queue.append(b'receive_bytes?name=defa')
        self.client_socket.recv_queue.append(b'ult&bytes=4096\n')
        data = b'0' * 4096
        self.client_socket.recv_queue.append(data[0:1000])
        self.client_socket.recv_queue.append(data[1000:])
        self.advance_time_and_run()
        self.receive_mock.assert_called_once_with(name="default", client=self._bcp_client, rawbytes=data)
        self.receive_mock.reset_mock()

        # test message without bytes
        self.receive_mock = MagicMock()
        self.client_socket.recv_queue.append(b'receive_bytes?param1=1&param2=2\n')
        self.advance_time_and_run()
        self.receive_mock.assert_called_once_with(param1="1", param2="2", client=self._bcp_client)
        self.receive_mock.reset_mock()

        # unknown method. should not crash
        self._bcp_client.send = MagicMock()
        self.client_socket.recv_queue.append(b'invalid_method?param1=1&param2=2\n')
        self.advance_time_and_run()


class TestBcpSocketMultipleClients(MpfTestCase):

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)

        # use bcp mock
        self.machine_config_patches['bcp'] = {}
        self.machine_config_patches['bcp']['servers'] = []
        self.receive_mock = None

    def get_use_bcp(self):
        return True

    def setUp(self):
        super().setUp()
        self._bcp_client_1 = self.machine.bcp.transport.get_named_client("local_display")
        self._bcp_client_2 = self.machine.bcp.transport.get_named_client("another_display")

    def _mock_loop(self):
        self.client_socket_1 = MockBcpQueueSocket(self.loop)
        self.clock.mock_socket("localhost", 5050, self.client_socket_1)
        self.client_socket_2 = MockBcpQueueSocket(self.loop)
        self.clock.mock_socket("localhost", 9001, self.client_socket_2)

    def get_config_file(self):
        return 'multiple_connections_config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/bcp/'

    async def receive_func(self, *args, **kwargs):
        self.receive_mock(*args, **kwargs)

    def testReceiveMessages(self):
        # test message without bytes
        self.receive_mock = MagicMock()
        self.machine.bcp.interface.register_command_callback("receive_msg", self.receive_func)

        self.client_socket_1.recv_queue.append(b'receive_msg?param1=1&param2=2\n')
        self.advance_time_and_run()
        self.receive_mock.assert_called_once_with(param1="1", param2="2", client=self._bcp_client_1)
        self.receive_mock.reset_mock()

        self.client_socket_2.recv_queue.append(b'receive_msg?param1=1&param2=2\n')
        self.advance_time_and_run()
        self.receive_mock.assert_called_once_with(param1="1", param2="2", client=self._bcp_client_2)



from mpf.tests.MpfTestCase import MpfTestCase


class TestI2cServoController(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/i2c_servo_controller/'

    def get_platform(self):
        return False

    def test_servo_move(self):
        self.assertEqual("64", self.machine.servos["servo1"].hw_servo.i2c_device.number)
        self.assertEqual(3, self.machine.servos["servo1"].hw_servo.number)
        self.assertEqual("bus1-64", self.machine.servos["servo2"].hw_servo.i2c_device.number)
        self.assertEqual(7, self.machine.servos["servo2"].hw_servo.number)
        self.assertEqual("64", self.machine.servos["servo3"].hw_servo.i2c_device.number)
        self.assertEqual(4, self.machine.servos["servo3"].hw_servo.number)
        self.assertEqual(self.machine.servos["servo1"].hw_servo.i2c_device,
                         self.machine.servos["servo3"].hw_servo.i2c_device)

        # move servo1
        self.assertEqual(119, self.machine.servos["servo1"].hw_servo.i2c_device.data[0x08 + 3 * 4])
        self.assertEqual(1, self.machine.servos["servo1"].hw_servo.i2c_device.data[0x09 + 3 * 4])
        self.machine.servos["servo1"].go_to_position(1.0)
        self.assertEqual(88, self.machine.servos["servo1"].hw_servo.i2c_device.data[0x08 + 3 * 4])
        self.assertEqual(2, self.machine.servos["servo1"].hw_servo.i2c_device.data[0x09 + 3 * 4])

        # move servo2
        self.assertEqual(119, self.machine.servos["servo2"].hw_servo.i2c_device.data[0x08 + 7 * 4])
        self.assertEqual(1, self.machine.servos["servo2"].hw_servo.i2c_device.data[0x09 + 7 * 4])
        self.machine.servos["servo2"].go_to_position(1.0)
        self.assertEqual(88, self.machine.servos["servo2"].hw_servo.i2c_device.data[0x08 + 7 * 4])
        self.assertEqual(2, self.machine.servos["servo2"].hw_servo.i2c_device.data[0x09 + 7 * 4])



"""Base class for ball device handlers."""
import asyncio

from mpf.core.utility_functions import Util


class BallDeviceStateHandler:

    """Base class for ball device handler."""

    __slots__ = ["ball_device", "machine", "_task"]

    def __init__(self, ball_device):
        """Initialise handler.

        Args:
        ----
            ball_device(mpf.devices.ball_device.ball_device.BallDevice): parent ball device
        """
        self.ball_device = ball_device
        self.machine = ball_device.machine
        self._task = None

    def stop(self):
        """Stop handler."""
        if self._task:
            self._task.cancel()
            try:
                self.machine.clock.loop.run_until_complete(self._task)
            except asyncio.CancelledError:
                pass

    def debug_log(self, *args, **kwargs):
        """Debug log."""
        self.ball_device.debug_log(*args, **kwargs)

    def info_log(self, *args, **kwargs):
        """Info log."""
        self.ball_device.info_log(*args, **kwargs)

    def warning_log(self, *args, **kwargs):
        """Warning log."""
        self.ball_device.warning_log(*args, **kwargs)

    async def initialise(self):
        """Initialise handler."""
        self._task = self.machine.clock.loop.create_task(self._run())
        self._task.add_done_callback(Util.raise_exceptions)

    async def _run(self):
        raise NotImplementedError()



"""File interface for pickled binary files."""
from typing import List

import pickle

from mpf.core.file_interface import FileInterface


class PickleInterface(FileInterface):

    """Loads and saves pickled binary files."""

    __slots__ = []  # type: List[str]

    file_types = ['.bin']

    def load(self, filename, expected_version_str=None, halt_on_error=True):
        """Load data from binary file."""
        del expected_version_str
        self.log.info("Loading %s", filename)
        try:
            with open(filename, "rb") as f:
                return pickle.load(f)
        except Exception:   # pylint: disable-msg=broad-except
            if halt_on_error:
                raise
            return None

    def save(self, filename, data):
        """Save data to binary file."""
        with open(filename, "wb") as f:
            pickle.dump(data, f)



"""Contains the TextUI class."""
from collections import defaultdict

from datetime import datetime
from psutil import cpu_percent, virtual_memory, Process

import mpf._version
from mpf.core.delays import DelayManager
from mpf.core.mpf_controller import MpfController

try:
    from asciimatics.scene import Scene
    from asciimatics.widgets import Frame, Layout, THEMES, Label, Divider, PopUpDialog, Widget
    from asciimatics.screen import Screen
except ImportError:
    Scene = None
    Frame = None
    Layout = object
    THEMES = None
    Label = None
    Divider = None
    PopUpDialog = None
    Widget = None
    Screen = None

MYPY = False
if MYPY:   # pragma: no cover
    from typing import List, Tuple, Dict  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import,ungrouped-imports
    from mpf.devices.ball_device.ball_device \
        import BallDevice  # pylint: disable-msg=cyclic-import,unused-import,ungrouped-imports
    from mpf.devices.switch import Switch   # pylint: disable-msg=cyclic-import,unused-import,ungrouped-imports


class MpfLayout(Layout):

    """Add clear function."""

    def __init__(self, columns, fill_frame=False):
        """Store max_height."""
        self._columns = []
        super().__init__(columns, fill_frame)
        self.max_height = None

    def clear_columns(self):
        """Clear all columns."""
        self._columns = [[] for _ in self._columns]

    def set_max_height(self, max_height):
        """Set max height."""
        self.max_height = max_height

    def fix(self, start_x, start_y, max_width, max_height):
        """Limit height."""
        if self.max_height:
            return min(super().fix(start_x, start_y, max_width, max_height), self.max_height)

        return super().fix(start_x, start_y, max_width, max_height)


# pylint: disable-msg=too-many-instance-attributes
class TextUi(MpfController):

    """Handles the text-based UI."""

    config_name = "text_ui"

    __slots__ = ["start_time", "machine", "_tick_task", "screen", "mpf_process", "ball_devices", "switches",
                 "config", "_pending_bcp_connection", "_asset_percent", "_player_widgets", "_machine_widgets",
                 "_bcp_status", "frame", "layout", "scene", "footer_memory", "switch_widgets", "mode_widgets",
                 "ball_device_widgets", "footer_cpu", "footer_mc_cpu", "footer_uptime", "delay", "_layout_change"]

    def __init__(self, machine: "MachineController") -> None:
        """Initialize TextUi."""
        super().__init__(machine)
        self.delay = DelayManager(machine)
        self.config = machine.config.get('text_ui', {})

        self.screen = None

        if not machine.options['text_ui'] or not Scene:
            return

        # hack to add themes until https://github.com/peterbrittain/asciimatics/issues/207 is implemented
        THEMES["mpf_theme"] = defaultdict(
            lambda: (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
            {
                "active_switch": (Screen.COLOUR_BLACK, Screen.A_NORMAL, Screen.COLOUR_GREEN),
                "pf_active": (Screen.COLOUR_GREEN, Screen.A_NORMAL, Screen.COLOUR_BLACK),
                "pf_inactive": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
                "label": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_BLACK),
                "title": (Screen.COLOUR_WHITE, Screen.A_NORMAL, Screen.COLOUR_RED),
                "title_exit": (Screen.COLOUR_BLACK, Screen.A_NORMAL, Screen.COLOUR_RED),
                "footer_cpu": (Screen.COLOUR_CYAN, Screen.A_NORMAL, Screen.COLOUR_BLACK),
                "footer_path": (Screen.COLOUR_YELLOW, Screen.A_NORMAL, Screen.COLOUR_BLACK),
                "footer_memory": (Screen.COLOUR_GREEN, Screen.A_NORMAL, Screen.COLOUR_BLACK),
                "footer_mc_cpu": (Screen.COLOUR_MAGENTA, Screen.A_NORMAL, Screen.COLOUR_BLACK),
            })

        self.start_time = datetime.now()
        self.machine = machine

        self.mpf_process = Process()
        self.ball_devices = list()      # type: List[BallDevice]

        self.switches = {}      # type: Dict[str, Switch]

        self.machine.events.add_handler('init_phase_2', self._init)
        # self.machine.events.add_handler('init_phase_3', self._init2)
        self.machine.events.add_handler('loading_assets',
                                        self._asset_load_change)
        self.machine.events.add_handler('bcp_connection_attempt',
                                        self._bcp_connection_attempt)
        self.machine.events.add_handler('asset_loading_complete',
                                        self._asset_load_complete)
        self.machine.events.add_handler('bcp_clients_connected',
                                        self._bcp_connected)
        self.machine.events.add_handler('shutdown', self.stop)
        self.machine.add_crash_handler(self.stop)
        self.machine.events.add_handler('player_number', self._update_player)
        self.machine.events.add_handler('player_ball', self._update_player)
        self.machine.events.add_handler('player_score', self._update_player)
        self.machine.events.add_handler('ball_ended',
                                        self._update_player)

        self._pending_bcp_connection = False
        self._asset_percent = 0
        self._bcp_status = (0, 0, 0)    # type: Tuple[float, int, int]
        self.switch_widgets = []        # type: List[Widget]
        self.mode_widgets = []          # type: List[Widget]
        self.ball_device_widgets = []   # type: List[Widget]
        self._machine_widgets = []      # type: List[Widget]
        self._player_widgets = []       # type: List[Widget]
        self.footer_memory = None
        self.footer_cpu = None
        self.footer_mc_cpu = None
        self.footer_uptime = None
        self._layout_change = True

        self._tick_task = self.machine.clock.schedule_interval(self._tick, 1)
        self._create_window()
        self._draw_screen()

    def _init(self, **kwargs):
        del kwargs
        for mode in self.machine.modes.values():
            self.machine.events.add_handler("mode_{}_started".format(mode.name), self._mode_change)
            self.machine.events.add_handler("mode_{}_stopped".format(mode.name), self._mode_change)

        self.machine.switch_controller.add_monitor(self._update_switches)
        self.machine.register_monitor("machine_vars", self._update_machine_vars)
        self.machine.variables.machine_var_monitor = True
        self.machine.bcp.interface.register_command_callback(
            "status_report", self._bcp_status_report)

        for bd in [x for x in self.machine.ball_devices.values() if not x.is_playfield()]:
            self.ball_devices.append(bd)

        self.ball_devices.sort()

        self._update_switch_layout()
        self._schedule_draw_screen()

    async def _bcp_status_report(self, client, cpu, rss, vms):
        del client
        self._bcp_status = cpu, rss, vms

    def _update_stats(self):
        # Runtime
        rt = (datetime.now() - self.start_time)
        mins, sec = divmod(rt.seconds + rt.days * 86400, 60)
        hours, mins = divmod(mins, 60)
        self.footer_uptime.text = 'RUNNING {:d}:{:02d}:{:02d}'.format(hours, mins, sec)

        # System Stats
        self.footer_memory.text = 'Free Memory (MB): {} CPU:{:3d}%'.format(
            round(virtual_memory().available / 1048576),
            round(cpu_percent(interval=None, percpu=False)))

        # MPF process stats
        self.footer_cpu.text = 'MPF (CPU RSS/VMS): {}% {}/{} MB    '.format(
            round(self.mpf_process.cpu_percent()),
            round(self.mpf_process.memory_info().rss / 1048576),
            round(self.mpf_process.memory_info().vms / 1048576))

        # MC process stats
        if self._bcp_status != (0, 0, 0):
            self.footer_mc_cpu.text = 'MC (CPU RSS/VMS) {}% {}/{} MB '.format(
                round(self._bcp_status[0]),
                round(self._bcp_status[1] / 1048576),
                round(self._bcp_status[2] / 1048576))
        else:
            self.footer_mc_cpu.text = ""

    def _update_switch_layout(self):
        num = 0
        self.switch_widgets = []
        self.switches = {}
        self.switch_widgets.append((Label("SWITCHES"), 1))
        self.switch_widgets.append((Divider(), 1))
        self.switch_widgets.append((Label(""), 2))
        self.switch_widgets.append((Divider(), 2))

        for sw in sorted(self.machine.switches.values()):
            if sw.invert:
                name = sw.name + '*'
            else:
                name = sw.name

            col = 1 if num <= int(len(self.machine.switches) / 2) else 2

            switch_widget = Label(name)
            if sw.state:
                switch_widget.custom_colour = "active_switch"

            self.switch_widgets.append((switch_widget, col))
            self.switches[sw.name] = (sw, switch_widget)

            num += 1

        self._schedule_draw_screen()

    def _update_switches(self, change, *args, **kwargs):
        del args
        del kwargs
        try:
            sw, switch_widget = self.switches[change.name]
        except KeyError:
            return
        if sw.state:
            switch_widget.custom_colour = "active_switch"
        else:
            switch_widget.custom_colour = "label"

        self._schedule_draw_screen()

    def _draw_switches(self):
        """Draw all switches."""
        for widget, column in self.switch_widgets:
            self.layout.add_widget(widget, column)

    def _mode_change(self, *args, **kwargs):
        # Have to call this on the next frame since the mode controller's
        # active list isn't updated yet
        del args
        del kwargs
        self.mode_widgets = []
        self.mode_widgets.append(Label("ACTIVE MODES"))
        self.mode_widgets.append(Divider())
        try:
            modes = self.machine.mode_controller.active_modes
        except AttributeError:
            modes = None

        if modes:
            for mode in modes:
                self.mode_widgets.append(Label('{} ({})'.format(mode.name, mode.priority)))
        else:
            self.mode_widgets.append(Label("No active modes"))

        # empty line at the end
        self.mode_widgets.append(Label(""))

        self._layout_change = True
        self._schedule_draw_screen()

    def _draw_modes(self):
        for widget in self.mode_widgets:
            self.layout.add_widget(widget, 0)

    def _draw_ball_devices(self):
        for widget in self.ball_device_widgets:
            self.layout.add_widget(widget, 3)

    def _update_ball_devices(self, **kwargs):
        del kwargs
        # TODO: do not create widgets. just update contents
        self.ball_device_widgets = []
        self.ball_device_widgets.append(Label("BALL COUNTS"))
        self.ball_device_widgets.append(Divider())

        try:
            for pf in self.machine.playfields.values():
                widget = Label('{}: {} '.format(pf.name, pf.balls))
                if pf.balls:
                    widget.custom_colour = "pf_active"
                else:
                    widget.custom_colour = "pf_inactive"
                self.ball_device_widgets.append(widget)

        except AttributeError:
            pass

        for bd in self.ball_devices:
            widget = Label('{}: {} ({})'.format(bd.name, bd.balls, bd.state))
            if bd.balls:
                widget.custom_colour = "pf_active"
            else:
                widget.custom_colour = "pf_inactive"

            self.ball_device_widgets.append(widget)

        self.ball_device_widgets.append(Label(""))

        self._layout_change = True
        self._schedule_draw_screen()

    def _update_player(self, **kwargs):
        del kwargs
        self._player_widgets = []
        self._player_widgets.append(Label("CURRENT PLAYER"))
        self._player_widgets.append(Divider())

        try:
            player = self.machine.game.player
            self._player_widgets.append(Label('PLAYER: {}'.format(player.number)))
            self._player_widgets.append(Label('BALL: {}'.format(player.ball)))
            self._player_widgets.append(Label('SCORE: {:,}'.format(player.score)))
        except AttributeError:
            self._player_widgets.append(Label("NO GAME IN PROGRESS"))
            return

        player_vars = player.vars.copy()
        player_vars.pop('score', None)
        player_vars.pop('number', None)
        player_vars.pop('ball', None)

        names = self.config.get('player_vars', player_vars.keys())
        for name in names:
            self._player_widgets.append(Label("{}: {}".format(name, player_vars[name])))

        self._layout_change = True
        self._schedule_draw_screen()

    def _draw_player(self, **kwargs):
        del kwargs
        for widget in self._player_widgets:
            self.layout.add_widget(widget, 3)

    def _update_machine_vars(self, **kwargs):
        """Update machine vars."""
        del kwargs
        self._machine_widgets = []
        self._machine_widgets.append(Label("MACHINE VARIABLES"))
        self._machine_widgets.append(Divider())
        machine_vars = self.machine.variables.machine_vars
        # If config defines explict vars to show, only show those. Otherwise, all
        names = self.config.get('machine_vars', machine_vars.keys())
        for name in names:
            self._machine_widgets.append(Label("{}: {}".format(name, machine_vars[name]['value'])))
        self._layout_change = True
        self._schedule_draw_screen()

    def _draw_machine_variables(self):
        """Draw machine vars."""
        for widget in self._machine_widgets:
            self.layout.add_widget(widget, 0)

    def _create_window(self):
        self.screen = Screen.open()
        self.frame = Frame(self.screen, self.screen.height, self.screen.width, has_border=False, title="Test")
        self.frame.set_theme("mpf_theme")

        title_layout = Layout([1, 5, 1])
        self.frame.add_layout(title_layout)

        title_left = Label("")
        title_left.custom_colour = "title"
        title_layout.add_widget(title_left, 0)

        title = 'Mission Pinball Framework v{}'.format(mpf._version.__version__)    # noqa
        title_text = Label(title, align="^")
        title_text.custom_colour = "title"
        title_layout.add_widget(title_text, 1)

        exit_label = Label("< CTRL + C > TO EXIT", align=">")
        exit_label.custom_colour = "title_exit"

        title_layout.add_widget(exit_label, 2)

        self.layout = MpfLayout([1, 1, 1, 1], fill_frame=True)
        self.frame.add_layout(self.layout)

        footer_layout = Layout([1, 1, 1])
        self.frame.add_layout(footer_layout)
        self.footer_memory = Label("", align=">")
        self.footer_memory.custom_colour = "footer_memory"
        self.footer_uptime = Label("", align=">")
        self.footer_uptime.custom_colour = "footer_memory"
        self.footer_mc_cpu = Label("")
        self.footer_mc_cpu.custom_colour = "footer_mc_cpu"
        self.footer_cpu = Label("")
        self.footer_cpu.custom_colour = "footer_cpu"
        footer_path = Label(self.machine.machine_path)
        footer_path.custom_colour = "footer_path"
        footer_empty = Label("")
        footer_empty.custom_colour = "footer_memory"

        footer_layout.add_widget(footer_path, 0)
        footer_layout.add_widget(self.footer_cpu, 0)
        footer_layout.add_widget(footer_empty, 1)
        footer_layout.add_widget(self.footer_mc_cpu, 1)
        footer_layout.add_widget(self.footer_uptime, 2)
        footer_layout.add_widget(self.footer_memory, 2)

        self.scene = Scene([self.frame], -1)
        self.screen.set_scenes([self.scene], start_scene=self.scene)

        # prevent main from scrolling out the footer
        self.layout.set_max_height(self.screen.height - 2)

    def _schedule_draw_screen(self):
        # schedule the draw in 10ms if it is not scheduled
        self.delay.add_if_doesnt_exist(10, self._draw_screen, "draw_screen")

    def _draw_screen(self):
        if not self.screen:
            # probably drawing during game end
            return

        if self._layout_change:
            self.layout.clear_columns()
            self._draw_modes()
            self._draw_machine_variables()
            self._draw_switches()
            self._draw_ball_devices()
            self._draw_player()
            self.frame.fix()
            self._layout_change = False

        self.screen.force_update()
        self.screen.draw_next_frame()

    def _tick(self):
        if self.screen.has_resized():
            self._create_window()

        self._update_ball_devices()
        self._update_stats()

        self._schedule_draw_screen()

        self.machine.bcp.transport.send_to_clients_with_handler(handler="_status_request",
                                                                bcp_command="status_request")

    def _bcp_connection_attempt(self, name, host, port, **kwargs):
        del name
        del kwargs
        self._pending_bcp_connection = PopUpDialog(self.screen,
                                                   'WAITING FOR MEDIA CONTROLLER {}:{}'.format(host, port), [])
        self.scene.add_effect(self._pending_bcp_connection)
        self._schedule_draw_screen()

    def _bcp_connected(self, **kwargs):
        del kwargs
        self.scene.remove_effect(self._pending_bcp_connection)
        self._schedule_draw_screen()

    def _asset_load_change(self, percent, **kwargs):
        del kwargs
        if self._asset_percent:
            self.scene.remove_effect(self._asset_percent)
        self._asset_percent = PopUpDialog(self.screen, 'LOADING ASSETS: {}%'.format(percent), [])
        self.scene.add_effect(self._asset_percent)
        self._schedule_draw_screen()

    def _asset_load_complete(self, **kwargs):
        del kwargs
        self.scene.remove_effect(self._asset_percent)
        self._schedule_draw_screen()

    def stop(self, **kwargs):
        """Stop the Text UI and restore the original console screen."""
        del kwargs
        if self.screen:
            self.machine.clock.unschedule(self._tick_task)
            self.screen.close(True)
            self.screen = None



from mpf.tests.MpfTestCase import MpfTestCase
from unittest.mock import MagicMock


class TestDeviceDriver(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/device/'

    def get_platform(self):
        return 'smart_virtual'

    def testBasicFunctions(self):
        # Make sure hardware devices have been configured for tests
        self.assertIn('flasher_01', self.machine.lights)

        # Setup platform function mock to test coil
        self.machine.coils["flasher_01"].hw_driver.disable = MagicMock()
        self.machine.coils["flasher_01"].hw_driver.enable = MagicMock()
        self.machine.coils["flasher_01"].hw_driver.pulse = MagicMock()

        # Flash
        self.post_event("flash")
        self.advance_time_and_run(.05)
        self.assertTrue(self.machine.coils["flasher_01"].hw_driver.enable.called)
        self.assertFalse(self.machine.coils["flasher_01"].hw_driver.disable.called)
        self.advance_time_and_run(.1)
        self.assertTrue(self.machine.coils["flasher_01"].hw_driver.disable.called)

        # Flash with placeholder
        self.machine.coils["flasher_01"].hw_driver.disable = MagicMock()
        self.machine.coils["flasher_01"].hw_driver.enable = MagicMock()
        self.machine.coils["flasher_01"].hw_driver.pulse = MagicMock()
        self.machine.coils["flasher_02"].hw_driver.disable = MagicMock()
        self.machine.coils["flasher_02"].hw_driver.enable = MagicMock()
        self.machine.coils["flasher_02"].hw_driver.pulse = MagicMock()

        self.post_event("flash2")
        self.advance_time_and_run(.05)
        self.assertTrue(self.machine.coils["flasher_01"].hw_driver.enable.called)
        self.assertFalse(self.machine.coils["flasher_01"].hw_driver.disable.called)
        self.assertTrue(self.machine.coils["flasher_02"].hw_driver.enable.called)
        self.assertFalse(self.machine.coils["flasher_02"].hw_driver.disable.called)
        self.advance_time_and_run(.1)
        self.assertTrue(self.machine.coils["flasher_01"].hw_driver.disable.called)
        self.assertTrue(self.machine.coils["flasher_02"].hw_driver.disable.called)



"""Random event config player."""
from mpf.core.config_player import ConfigPlayer
from mpf.core.randomizer import Randomizer
from mpf.core.utility_functions import Util


class RandomEventPlayer(ConfigPlayer):

    """Plays a random event based on config."""

    config_file_section = 'random_event_player'
    show_section = 'random_events'

    __slots__ = ["_machine_wide_dict"]

    def __init__(self, machine):
        """Initialise random event player."""
        super().__init__(machine)
        self._machine_wide_dict = {}

    @staticmethod
    def is_entry_valid_outside_mode(settings) -> bool:
        """Return true if scope is not player."""
        return settings['scope'] != "player"

    def _get_randomizer(self, settings, context, calling_context):
        key = "random_{}.{}".format(context, calling_context)
        if settings['scope'] == "player":
            if not self.machine.game.player[key]:
                self.machine.game.player[key] = Randomizer(
                    settings['events'], self.machine, template_type="event")
                r'''player_var: random_(x).(y)

                desc: Holds references to Randomizer settings that need to be
                tracked on a player basis. There is nothing you need to know
                or do with this, rather this is just FYI on what the player
                variables that start with "random\_" are.
                '''

                if settings['force_all']:
                    self.machine.game.player[key].force_all = True

                if not settings['force_different']:
                    self.machine.game.player[key].force_different = False

                if settings['disable_random']:
                    self.machine.game.player[key].disable_random = True

                self.machine.game.player[key].fallback_value = settings.get('fallback_event')

            return self.machine.game.player[key]

        if key not in self._machine_wide_dict:
            self._machine_wide_dict[key] = Randomizer(
                settings['events'], self.machine, template_type="event")

            if settings['force_all']:
                self._machine_wide_dict[key].force_all = True

            if not settings['force_different']:
                self._machine_wide_dict[key].force_different = False

            if settings['disable_random']:
                self._machine_wide_dict[key].disable_random = True

            self._machine_wide_dict[key].fallback_value = settings.get('fallback_event')

        return self._machine_wide_dict[key]

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Play a random event from list based on config."""
        del priority
        randomizer = self._get_randomizer(settings, context, calling_context)
        # With conditional events in randomizer, there may not be a next event
        next_event = randomizer.get_next(kwargs)
        if next_event:
            self.machine.events.post(next_event, **kwargs)

    def validate_config_entry(self, settings, name):
        """Validate one entry of this player."""
        config = self._parse_config(settings, name)
        return config

    def get_express_config(self, value):
        """Parse express config."""
        return {"events": self.get_list_config(Util.string_to_event_list(value))}

    def get_list_config(self, value):
        """Parse list."""
        return {"events": value}



"""Light group devices to comfortably configure groups of lights."""
import abc
import copy

import math

from typing import List

from mpf.core.machine import MachineController

from mpf.core.system_wide_device import SystemWideDevice
from mpf.devices.light import Light


class LightGroup(SystemWideDevice):

    """An abstract group of lights."""

    __slots__ = ["lights"]

    def __init__(self, machine: MachineController, name) -> None:
        """Initialise light group."""
        super().__init__(machine, name)

        self.lights = []        # type: List[Light]

    @classmethod
    def prepare_config(cls, config: dict, is_mode_config: bool):
        """Add light_template and number to config.

        Args:
        ----
            config: unparsed config
            is_mode_config: if in mode (not used)
        """
        del is_mode_config
        if 'light_template' not in config:
            config['light_template'] = {}

        return config

    async def _initialize(self):
        await super()._initialize()
        self._create_lights()

        for light in self.lights:
            await light.device_added_system_wide()

    def get_token(self):
        """Return all lights in group as token."""
        return {'lights': self.lights}

    def _create_light_at_index(self, index, x, y, relative_index):
        light = Light(self.machine, "{}_light_{}".format(self.name, relative_index))
        tags = [self.name]
        tags.extend(self.config['tags'])
        light_config = copy.deepcopy(self.config['light_template'])
        if self.config['start_channel']:
            if relative_index == 0:
                light_config['start_channel'] = self.config['start_channel']
            else:
                light_config['previous'] = "{}_light_{}".format(self.name, relative_index - 1)
        elif self.config['number_template']:
            light_config['number'] = self.config['number_template'].format(index)
        else:
            light_config['number'] = index
        light_config['tags'].append(self.name)
        light_config['x'] = x
        light_config['y'] = y
        light_config = light.validate_and_parse_config(light_config, False)
        light.load_config(light_config)
        self.lights.append(light)
        self.machine.lights[light.name] = light

    @abc.abstractmethod
    def _create_lights(self):
        raise NotImplementedError("Implement")

    def color(self, color, fade_ms=None, priority=0, key=None):
        """Call color on all lights in this group."""
        for light in self.lights:
            light.color(color, fade_ms, priority, key)


class LightStrip(LightGroup):

    """A light stripe."""

    config_section = 'light_stripes'
    collection = 'light_stripes'
    class_label = 'light_stripe'

    __slots__ = []  # type: List[str]

    def _create_lights(self):
        distance = 0
        for index in range(self.config['number_start'], self.config['number_start'] + self.config['count']):
            if self.config['start_x'] is not None:
                x = self.config['start_x'] + math.sin(self.config['direction'] / 180 * math.pi) * distance
                y = self.config['start_y'] + math.cos(self.config['direction'] / 180 * math.pi) * distance
                distance += self.config['distance']
            else:
                x = y = None

            relative_index = index - self.config['number_start']
            self._create_light_at_index(index, x, y, relative_index)


class LightRing(LightGroup):

    """A light ring."""

    config_section = 'light_rings'
    collection = 'light_rings'
    class_label = 'light_ring'

    __slots__ = []  # type: List[str]

    def _create_lights(self):
        angle = self.config['start_angle'] / 180 * math.pi
        for index in range(self.config['number_start'], self.config['number_start'] + self.config['count']):
            if self.config['center_x'] is not None:
                x = self.config['center_x'] + math.sin(angle) * self.config['radius']
                y = self.config['center_y'] + math.cos(angle) * self.config['radius']
                angle += 2 * math.pi / self.config['count']
            else:
                x = y = None

            relative_index = index - self.config['number_start']
            self._create_light_at_index(index, x, y, relative_index)
